from __future__ import annotations

from itertools import chain as flatten
from math import prod
from typing import Any, Sequence, Tuple

import pyqtorch as pyq
import sympy
import torch
from pyqtorch.apply import apply_operator
from pyqtorch.embed import Embedding
from pyqtorch.matrices import _dagger
from pyqtorch.time_dependent.sesolve import sesolve
from pyqtorch.utils import is_diag
from torch import (
    Tensor,
    cdouble,
    diag_embed,
    diagonal,
    exp,
    float64,
    linalg,
    tensor,
    transpose,
)
from torch import device as torch_device
from torch import dtype as torch_dtype
from torch.nn import Module, ParameterDict

from qadence.backends.utils import (
    finitediff,
    pyqify,
    unpyqify,
)
from qadence.blocks import (
    AbstractBlock,
    AddBlock,
    ChainBlock,
    CompositeBlock,
    MatrixBlock,
    ParametricBlock,
    PrimitiveBlock,
    ScaleBlock,
    TimeEvolutionBlock,
)
from qadence.blocks.block_to_tensor import (
    _block_to_tensor_embedded,
    block_to_tensor,
)
from qadence.blocks.primitive import ProjectorBlock
from qadence.blocks.utils import parameters
from qadence.operations import (
    U,
    multi_qubit_gateset,
    non_unitary_gateset,
    single_qubit_gateset,
    three_qubit_gateset,
    two_qubit_gateset,
)
from qadence.types import OpName

from .config import Configuration

# Tdagger is not supported currently
supported_gates = list(set(OpName.list()) - set([OpName.TDAGGER]))
"""The set of supported gates.

Tdagger is currently not supported.
"""


def is_single_qubit_chain(block: AbstractBlock) -> bool:
    return (
        isinstance(block, (ChainBlock))
        and block.n_supports == 1
        and all([isinstance(b, (ParametricBlock, PrimitiveBlock)) for b in block])
        and not any([isinstance(b, (ScaleBlock, U)) for b in block])
    )


def convert_block(
    block: AbstractBlock, n_qubits: int = None, config: Configuration = None
) -> Sequence[Module | Tensor | str | sympy.Expr]:
    if isinstance(block, (Tensor, str, sympy.Expr)):  # case for hamevo generators
        if isinstance(block, Tensor):
            block = block.permute(1, 2, 0)  # put batch size in the back
        return [block]
    qubit_support = block.qubit_support
    if n_qubits is None:
        n_qubits = max(qubit_support) + 1

    if config is None:
        config = Configuration()

    if isinstance(block, ScaleBlock):
        scaled_ops = convert_block(block.block, n_qubits, config)
        scale = (
            tensor([block.parameters.parameter], dtype=float64)
            if not block.is_parametric
            else config.get_param_name(block)[0]
        )
        return [pyq.Scale(pyq.Sequence(scaled_ops), scale)]

    elif isinstance(block, TimeEvolutionBlock):
        # TODO add native pyq hamevo
        # generator = convert_block(block.generator, n_qubits, config)[0]  # type: ignore[arg-type]
        # time_param = config.get_param_name(block)[0]
        # is_parametric = (
        #     block.generator.is_parametric if isinstance(block.generator, AbstractBlock) else False
        # )
        # return [
        #     pyq.HamiltonianEvolution(
        #         qubit_support=qubit_support,
        #         generator=generator,
        #         time=time_param,
        #         generator_parametric=is_parametric,  # type: ignore[union-attr]
        #     )
        # ]
        return [PyQHamiltonianEvolution(qubit_support, n_qubits, block, config)]
    elif isinstance(block, MatrixBlock):
        return [pyq.primitive.Primitive(block.matrix, block.qubit_support)]
    elif isinstance(block, CompositeBlock):
        ops = list(flatten(*(convert_block(b, n_qubits, config) for b in block.blocks)))
        if isinstance(block, AddBlock):
            return [pyq.Add(ops)]  # add
        elif is_single_qubit_chain(block) and config.use_single_qubit_composition:
            return [pyq.Merge(ops)]  # for chains of single qubit ops on the same qubit
        else:
            return [pyq.Sequence(ops)]  # for kron and chain
    elif isinstance(block, tuple(non_unitary_gateset)):
        if isinstance(block, ProjectorBlock):
            projector = getattr(pyq, block.name)
            if block.name == OpName.N:
                return [projector(target=qubit_support)]
            else:
                return [projector(qubit_support=qubit_support, ket=block.ket, bra=block.bra)]
        else:
            return [getattr(pyq, block.name)(qubit_support[0])]
    elif isinstance(block, tuple(single_qubit_gateset)):
        pyq_cls = getattr(pyq, block.name)
        if isinstance(block, ParametricBlock):
            if isinstance(block, U):
                op = pyq_cls(qubit_support[0], *config.get_param_name(block))
            else:
                op = pyq_cls(qubit_support[0], config.get_param_name(block)[0])
        else:
            op = pyq_cls(qubit_support[0])
        return [op]
    elif isinstance(block, tuple(two_qubit_gateset)):
        pyq_cls = getattr(pyq, block.name)
        if isinstance(block, ParametricBlock):
            op = pyq_cls(qubit_support[0], qubit_support[1], config.get_param_name(block)[0])
        else:
            op = pyq_cls(qubit_support[0], qubit_support[1])
        return [op]
    elif isinstance(block, tuple(three_qubit_gateset) + tuple(multi_qubit_gateset)):
        block_name = block.name[1:] if block.name.startswith("M") else block.name
        pyq_cls = getattr(pyq, block_name)
        if isinstance(block, ParametricBlock):
            op = pyq_cls(qubit_support[:-1], qubit_support[-1], config.get_param_name(block)[0])
        else:
            if "CSWAP" in block_name:
                op = pyq_cls(qubit_support[:-2], qubit_support[-2:])
            else:
                op = pyq_cls(qubit_support[:-1], qubit_support[-1])
        return [op]
    else:
        raise NotImplementedError(
            f"Non supported operation of type {type(block)}. "
            "In case you are trying to run an `AnalogBlock`, make sure you "
            "specify the `device_specs` in your `Register` first."
        )


class PyQHamiltonianEvolution(Module):
    def __init__(
        self,
        qubit_support: Tuple[int, ...],
        n_qubits: int,
        block: TimeEvolutionBlock,
        config: Configuration,
    ):
        super().__init__()
        self.qubit_support = qubit_support
        self.n_qubits = n_qubits
        self.param_names = config.get_param_name(block)
        self.block = block
        self.hmat: Tensor
        self.config = config

        if isinstance(block.generator, AbstractBlock) and not block.generator.is_parametric:
            hmat = block_to_tensor(
                block.generator,
                qubit_support=self.qubit_support,
                use_full_support=False,
            )
            hmat = hmat.permute(1, 2, 0)
            self.register_buffer("hmat", hmat)
            self._hamiltonian = lambda self, values: self.hmat

        elif isinstance(block.generator, Tensor):
            m = block.generator.to(dtype=cdouble)
            hmat = block_to_tensor(
                MatrixBlock(
                    m,
                    qubit_support=block.qubit_support,
                    check_unitary=False,
                    check_hermitian=True,
                ),
                qubit_support=self.qubit_support,
                use_full_support=False,
            )
            hmat = hmat.permute(1, 2, 0)
            self.register_buffer("hmat", hmat)
            self._hamiltonian = lambda self, values: self.hmat

        elif isinstance(block.generator, sympy.Basic):
            self._hamiltonian = (
                lambda self, values: values[self.param_names[1]].squeeze(3).permute(1, 2, 0)
            )
            # FIXME Why are we squeezing
        else:

            def _hamiltonian(self: PyQHamiltonianEvolution, values: dict[str, Tensor]) -> Tensor:
                hmat = _block_to_tensor_embedded(
                    block.generator,  # type: ignore[arg-type]
                    values=values,
                    qubit_support=self.qubit_support,
                    use_full_support=False,
                    device=self.device,
                )
                return hmat.permute(1, 2, 0)

            self._hamiltonian = _hamiltonian

        self._time_evolution = lambda values: values[self.param_names[0]]
        self._device: torch_device = (
            self.hmat.device if hasattr(self, "hmat") else torch_device("cpu")
        )
        self._dtype: torch_dtype = self.hmat.dtype if hasattr(self, "hmat") else cdouble

    def _unitary(self, hamiltonian: Tensor, time_evolution: Tensor) -> Tensor:
        self.batch_size = max(hamiltonian.size()[2], len(time_evolution))
        diag_check = tensor(
            [is_diag(hamiltonian[..., i]) for i in range(hamiltonian.size()[2])], device=self.device
        )

        def _evolve_diag_operator(hamiltonian: Tensor, time_evolution: Tensor) -> Tensor:
            evol_operator = diagonal(hamiltonian) * (-1j * time_evolution).view((-1, 1))
            evol_operator = diag_embed(exp(evol_operator))
            return transpose(evol_operator, 0, -1)

        def _evolve_matrixexp_operator(hamiltonian: Tensor, time_evolution: Tensor) -> Tensor:
            evol_operator = transpose(hamiltonian, 0, -1) * (-1j * time_evolution).view((-1, 1, 1))
            evol_operator = linalg.matrix_exp(evol_operator)
            return transpose(evol_operator, 0, -1)

        evolve_operator = (
            _evolve_diag_operator if bool(prod(diag_check)) else _evolve_matrixexp_operator
        )
        return evolve_operator(hamiltonian, time_evolution)

    def unitary(self, values: dict[str, Tensor]) -> Tensor:
        """The evolved operator given current parameter values for generator and time evolution."""
        return self._unitary(self._hamiltonian(self, values), self._time_evolution(values))

    def jacobian_time(self, values: dict[str, Tensor]) -> Tensor:
        """Approximate jacobian of the evolved operator with respect to time evolution."""
        return finitediff(
            lambda t: self._unitary(time_evolution=t, hamiltonian=self._hamiltonian(self, values)),
            values[self.param_names[0]].reshape(-1, 1),
            (0,),
        )

    def jacobian_generator(self, values: dict[str, Tensor]) -> Tensor:
        """Approximate jacobian of the evolved operator with respect to generator parameter(s)."""
        if len(self.param_names) > 2:
            raise NotImplementedError(
                "jacobian_generator does not support generators\
                                        with more than 1 parameter."
            )

        def _generator(val: Tensor) -> Tensor:
            val_copy = values.copy()
            val_copy[self.param_names[1]] = val
            hmat = _block_to_tensor_embedded(
                self.block.generator,  # type: ignore[arg-type]
                values=val_copy,
                qubit_support=self.qubit_support,
                use_full_support=False,
                device=self.device,
            )
            return hmat.permute(1, 2, 0)

        return finitediff(
            lambda v: self._unitary(
                time_evolution=self._time_evolution(values), hamiltonian=_generator(v)
            ),
            values[self.param_names[1]].reshape(-1, 1),
            (0,),
        )

    def dagger(self, values: dict[str, Tensor]) -> Tensor:
        """Dagger of the evolved operator given the current parameter values."""
        return _dagger(self.unitary(values))

    def _get_time_parameter(self) -> str:
        # get unique time parameters
        unique_time_params = set()
        for p in parameters(self.block.generator):  # type: ignore [arg-type]
            if getattr(p, "is_time", False):
                unique_time_params.add(str(p))

        if len(unique_time_params) > 1:
            raise Exception("Only a single time parameter is supported.")

        return unique_time_params.pop()

    def forward(
        self,
        state: Tensor,
        values: dict[str, Tensor] | ParameterDict = dict(),
        embedding: Embedding | None = None,
    ) -> Tensor:
        if getattr(self.block.generator, "is_time_dependent", False):  # type: ignore [union-attr]

            def Ht(t: Tensor | float) -> Tensor:
                # values dict has to change with new value of t
                # initial value of a feature parameter inside generator block
                # has to be inferred here
                new_vals = dict()
                for str_expr, val in values.items():
                    expr = sympy.sympify(str_expr)
                    t_symb = sympy.Symbol(self._get_time_parameter())
                    free_symbols = expr.free_symbols
                    if t_symb in free_symbols:
                        # create substitution list for time and feature params
                        subs_list = [(t_symb, t)]

                        if len(free_symbols) > 1:
                            # get feature param symbols
                            feat_symbols = free_symbols.difference(set([t_symb]))

                            # get feature param values
                            feat_vals = values["orig_param_values"]

                            # update substitution list with feature param values
                            for fs in feat_symbols:
                                subs_list.append((fs, feat_vals[str(fs)]))

                        # evaluate expression with new time param value
                        new_vals[str_expr] = torch.tensor(float(expr.subs(subs_list)))
                    else:
                        # expression doesn't contain time parameter - copy it as is
                        new_vals[str_expr] = val

                # get matrix form of generator
                hmat = _block_to_tensor_embedded(
                    self.block.generator,  # type: ignore[arg-type]
                    values=new_vals,
                    qubit_support=self.qubit_support,
                    use_full_support=False,
                    device=self.device,
                ).squeeze(0)

                return hmat

            tsave = torch.linspace(0, self.block.duration, self.config.n_steps_hevo)  # type: ignore [attr-defined]
            result = pyqify(
                sesolve(Ht, unpyqify(state).T[:, 0:1], tsave, self.config.ode_solver).states[-1].T
            )
        else:
            result = apply_operator(
                state,
                self.unitary(values),
                self.qubit_support,
                self.n_qubits,
                self.batch_size,
            )

        return result

    @property
    def device(self) -> torch_device:
        return self._device

    @property
    def dtype(self) -> torch_dtype:
        return self._dtype

    def to(self, *args: Any, **kwargs: Any) -> PyQHamiltonianEvolution:
        if hasattr(self, "hmat"):
            self.hmat = self.hmat.to(*args, **kwargs)
            self._device = self.hmat.device
            self._dtype = self.hmat.dtype
        return self



from __future__ import annotations

from qadence.analog.hamiltonian_terms import (
    rydberg_drive_hamiltonian,
    rydberg_interaction_hamiltonian,
    rydberg_pattern_hamiltonian,
)
from qadence.blocks import chain
from qadence.blocks.abstract import AbstractBlock
from qadence.blocks.analog import (
    AnalogKron,
    ConstantAnalogRotation,
    InteractionBlock,
)
from qadence.circuit import QuantumCircuit
from qadence.operations import HamEvo
from qadence.register import Register
from qadence.transpile import apply_fn_to_blocks


def add_background_hamiltonian(
    circuit: QuantumCircuit | AbstractBlock,
    register: Register | None = None,
) -> QuantumCircuit | AbstractBlock:
    """
    Parses a `QuantumCircuit` to transform `AnalogBlocks` to `HamEvo`.

    Depends on the circuit `Register` and included `RydbergDevice` specifications.

    Currently checks if input is either circuit or block and adjusts
    the ouput accordingly. Running this function on single blocks is
    currently used for eigenvalue computation for GPSR.

    Arguments:
        circuit: the circuit to parse, or single block to transform.
        register: needed for calling the function on a single block.
    """
    # FIXME: revisit eigenvalues of analog blocks and clean this code.

    is_circuit_input = isinstance(circuit, QuantumCircuit)

    if not is_circuit_input and register is None:
        raise ValueError("Block input requires an input to the `register` argument.")

    input_block: AbstractBlock = circuit.block if is_circuit_input else circuit  # type: ignore
    input_register: Register = circuit.register if is_circuit_input else register  # type: ignore

    if input_register.device_specs is not None:
        # Create interaction hamiltonian:
        h_int = rydberg_interaction_hamiltonian(input_register)

        # Create addressing pattern:
        h_addr = rydberg_pattern_hamiltonian(input_register)

        output_block = apply_fn_to_blocks(
            input_block,
            _analog_to_hevo,
            input_register,
            h_int,
            h_addr,
        )
    else:
        output_block = input_block

    if is_circuit_input:
        return QuantumCircuit(input_register, output_block)
    else:
        return output_block


def _build_rot_ham_evo(
    block: ConstantAnalogRotation,
    h_int: AbstractBlock,
    h_addr: AbstractBlock | None,
    h_drive: AbstractBlock,
) -> HamEvo:
    h_block = h_int
    if block.add_pattern and h_addr is not None:
        h_block += h_addr
    duration = block.parameters.duration
    h_norm = block.parameters.h_norm
    h_block += h_drive
    return HamEvo(h_block / h_norm, duration * h_norm / 1000)


def _analog_to_hevo(
    block: AbstractBlock, register: Register, h_int: AbstractBlock, h_addr: AbstractBlock | None
) -> AbstractBlock:
    """
    Converter from AnalogBlock to the respective HamEvo.

    Any other block not covered by the specific conditions below is left unchanged.
    """

    if isinstance(block, InteractionBlock):
        h_background = h_int
        if block.add_pattern and h_addr is not None:
            h_background += h_addr
        duration = block.parameters.duration
        return HamEvo(h_background, duration / 1000)

    if isinstance(block, ConstantAnalogRotation):
        h_drive = rydberg_drive_hamiltonian(block, register)
        return _build_rot_ham_evo(block, h_int, h_addr, h_drive)

    if isinstance(block, AnalogKron):
        # Needed to ensure kronned Analog blocks are implemented
        # in sequence, consistent with the current Pulser implementation.
        # FIXME: Revisit this assumption and the need for AnalogKron to have
        # the same duration, and clean this code accordingly.
        # https://github.com/pasqal-io/qadence/issues/226
        ops = []
        for block in block.blocks:
            if isinstance(block, ConstantAnalogRotation):
                h_drive = rydberg_drive_hamiltonian(block, register)
                ops.append(_build_rot_ham_evo(block, h_int, h_addr, h_drive))
        if len(ops) == 0:
            ops.append(HamEvo(h_background, duration / 1000))  # type: ignore [arg-type]
        return chain(*ops)

    return block



from __future__ import annotations

import itertools
from enum import Enum
from logging import getLogger
from typing import Any, List, Tuple, Union

import sympy

from qadence.blocks import AbstractBlock
from qadence.blocks.utils import get_pauli_blocks, unroll_block_with_scaling
from qadence.parameters import Parameter, evaluate

# from qadence.types import TNumber, TParameter
from qadence.types import PI
from qadence.types import LTSOrder as Order

logger = getLogger(__name__)

# flatten a doubly-nested list
flatten = lambda a: list(itertools.chain(*a))  # noqa: E731


def change_to_z_basis(block: AbstractBlock, sign: int) -> list[AbstractBlock]:
    """A simple function to do basis transformation of blocks of X and Y.

    This needs to be generalized beyond 2 terms
    """

    # import here due to circular import issue
    from qadence.operations import RX, H, X, Y

    qubit = block.qubit_support[0]

    if isinstance(block, X):
        return [H(target=qubit)]

    elif isinstance(block, Y):
        return [RX(parameter=sign * PI / 2.0, target=qubit)]

    return []


def time_discretisation(
    parameter: Parameter, max_steps: int = 10
) -> Tuple[Union[float, complex, Any], int]:
    """Checks and returns a numerically stable.

    time step that is used in the product formula
    """
    # the approximation gives better results for t -> 0
    # the delta t needs to be numerically stable
    # ! constant time steps
    # ! the more steps, the more computationally expensive circuit
    # https://arxiv.org/pdf/1403.3469.pdf

    # check the time and log warning on duration if needed
    time = evaluate(parameter)

    if (time / max_steps) > 1e-3:  # type: ignore
        logger.warning(
            """Please consider running the H evolution for
             a shorter time to get a better approximation."""
        )

    t_delta = parameter / max_steps  # ! check numerical stability
    return t_delta, max_steps


def lie_trotter_suzuki(
    block: AbstractBlock | List, parameter: Parameter, order: Enum = Order.BASIC
) -> list[AbstractBlock]:
    # get the block and transform it to a list of blocks
    # do the correct decomposition
    # return a list of blocks

    if not isinstance(block, list):
        block_list = unroll_block_with_scaling(block)
    else:  # recursive for 4th order
        block_list = block

    if order == Order.BASIC:  # Lie-Trotter 1st order
        return decompose_pauli_exp(block_list, parameter)

    else:  # Suzuki-Trotter 2nd and 4th order
        # ! handle time properly, break up into small time steps
        # get a useful numerically stable time step

        t_delta, t_steps = time_discretisation(parameter)

        if order == Order.ST2:  # Suzuki-Trotter 2nd order
            outer = decompose_pauli_exp(block_list[:-1], Parameter(t_delta / 2.0))
            inner = decompose_pauli_exp([block_list[-1]], Parameter(t_delta))
            return (outer + inner + list(reversed(outer))) * t_steps

        else:  # Suzuki-Trotter 4th order
            p2 = (4 - 4 ** (1 / 3)) ** -1  # minimises the 'ideal' error in the recursive formula
            outer = lie_trotter_suzuki(block_list, Parameter((t_delta * p2)), order=Order.ST2)
            inner = lie_trotter_suzuki(
                block_list, Parameter((1 - 4 * p2) * t_delta), order=Order.ST2
            )
            return (2 * outer + inner + outer * 2) * t_steps


def decompose_pauli_exp(block_list: list, parameter: Parameter | sympy.Expr) -> list[AbstractBlock]:
    """A simple function to do decompositions of Pauli exponential operators into digital gates."""

    # import here due to circular import issue
    from qadence.operations import CNOT, RZ

    blocks = []

    for bl, scale in block_list:
        # extract Pauli operations and raise an error in case
        # a non-Pauli operation is found since it cannot be
        # decomposed
        n_blocks = len(get_pauli_blocks(bl))

        # ensure that we keep the parameter as trainable
        fact = 2.0 * parameter * scale

        blist: list[AbstractBlock] = bl if n_blocks >= 2 else [bl]  # type: ignore[assignment]
        indx = [b.qubit_support[0] for b in blist]
        ztarget = max(indx)

        cnot_sequence = [CNOT(i, i + 1) for i in range(min(indx), max(indx))]
        basis_fwd = [change_to_z_basis(blist[i], 1) for i in range(len(blist))]
        rot = [RZ(parameter=Parameter(fact), target=ztarget)]
        basis_bkd = [change_to_z_basis(blist[i], -1) for i in range(len(blist) - 1, -1, -1)]

        # NOTE
        # perform the following operations in sequence to perform the decomposition of a
        # polynomial Pauli term, for more details, see: https://arxiv.org/abs/1001.3855
        # - change to Z basis for all the needed qubit operators
        # - apply a CNOT ladder on the full qubit support where operators are acting
        # - apply a RZ rotation on the last qubit
        # - apply the reverse CNOT ladder
        # - go back to the original basis
        blocks.extend(
            flatten(basis_fwd)
            + cnot_sequence
            + rot
            + list(reversed(cnot_sequence))
            + flatten(basis_bkd)
        )

    return blocks



from __future__ import annotations

from .apply_fn import apply_fn_to_blocks
from .block import (
    chain_single_qubit_ops,
    repeat,
    scale_primitive_blocks_only,
    set_trainable,
    validate,
)
from .circuit import fill_identities
from .digitalize import digitalize
from .flatten import flatten
from .invert import invert_endianness, reassign
from .transpile import blockfn_to_circfn, transpile

__all__ = ["set_trainable", "invert_endianness"]



from __future__ import annotations

from collections import Counter
from functools import reduce

import numpy as np
import numpy.typing as npt
import torch
from numpy.linalg import inv, matrix_rank, pinv
from scipy.linalg import norm
from scipy.optimize import LinearConstraint, minimize

from qadence.mitigations.protocols import Mitigations
from qadence.noise.protocols import Noise
from qadence.types import ReadOutOptimization


def corrected_probas(p_corr: npt.NDArray, T: npt.NDArray, p_raw: npt.NDArray) -> np.double:
    return norm(T @ p_corr.T - p_raw.T, ord=2) ** 2


def mle_solve(p_raw: npt.NDArray) -> npt.NDArray:
    """
    Compute the MLE probability vector.

    Algorithmic details can be found in https://arxiv.org/pdf/1106.5458.pdf Page(3).
    """
    # Sort p_raw by values while keeping track of indices.
    index_sort = p_raw.argsort()
    p_sort = p_raw[index_sort]
    neg_sum = 0
    breakpoint = len(p_sort) - 1

    for i in range(len(p_sort)):
        ## if neg_sum cannot be distributed among other probabilities, continue to accumulate
        if p_sort[i] + neg_sum / (len(p_sort) - i) < 0:
            neg_sum += p_sort[i]
            p_sort[i] = 0
        # set breakpoint to current index
        else:
            breakpoint = i
            break
    ## number of entries to which i can distribute(includes breakpoint)
    size = len(p_sort) - breakpoint
    p_sort[breakpoint:] += neg_sum / size

    re_index_sort = index_sort.argsort()
    p_corr = p_sort[re_index_sort]

    return p_corr


def renormalize_counts(corrected_counts: npt.NDArray, n_shots: int) -> npt.NDArray:
    """Renormalize counts rounding discrepancies."""
    total_counts = sum(corrected_counts)
    if total_counts != n_shots:
        counts_diff = total_counts - n_shots
        corrected_counts -= counts_diff
        corrected_counts = np.where(corrected_counts < 0, 0, corrected_counts)
        sum_corrected_counts = sum(corrected_counts)

        renormalization_factor = n_shots / sum_corrected_counts
        corrected_counts = np.rint(corrected_counts * renormalization_factor).astype(int)
    return corrected_counts


def matrix_inv(K: npt.NDArray) -> npt.NDArray:
    return inv(K) if matrix_rank(K) == K.shape[0] else pinv(K)


def mitigation_minimization(
    noise: Noise,
    mitigation: Mitigations,
    samples: list[Counter],
) -> list[Counter]:
    """Minimize a correction matrix subjected to stochasticity constraints.

    See Equation (5) in https://arxiv.org/pdf/2001.09980.pdf.
    See Page(3) in https://arxiv.org/pdf/1106.5458.pdf for MLE implementation

    Args:
        noise: Specifies confusion matrix and default error probability
        mitigation: Selects additional mitigation options based on noise choice.
        For readout we have the following mitigation options for optimization
        1. constrained 2. mle. Default : mle
        samples: List of samples to be mitigated

    Returns:
        Mitigated counts computed by the algorithm
    """
    noise_matrices = noise.options.get("noise_matrix", noise.options["confusion_matrices"])
    optimization_type = mitigation.options.get("optimization_type", ReadOutOptimization.MLE)
    n_qubits = len(list(samples[0].keys())[0])
    n_shots = sum(samples[0].values())
    corrected_counters: list[Counter] = []

    if optimization_type == ReadOutOptimization.CONSTRAINED:
        # Build the whole T matrix.
        T_matrix = reduce(torch.kron, noise_matrices).detach().numpy()

    if optimization_type == ReadOutOptimization.MLE:
        # Check if matrix is singular and use appropriate inverse.
        noise_matrices_inv = list(map(matrix_inv, noise_matrices.numpy()))
        T_inv = reduce(np.kron, noise_matrices_inv)

    for sample in samples:
        bitstring_length = 2**n_qubits
        # List of bitstrings in lexicographical order.
        ordered_bitstrings = [f"{i:0{n_qubits}b}" for i in range(bitstring_length)]
        # Array of raw probabilites.
        p_raw = np.array([sample[bs] for bs in ordered_bitstrings]) / n_shots

        if optimization_type == ReadOutOptimization.CONSTRAINED:
            # Initial random guess in [0,1].
            p_corr0 = np.random.rand(bitstring_length)
            # Stochasticity constraints.
            normality_constraint = LinearConstraint(
                np.ones(bitstring_length).astype(int), lb=1.0, ub=1.0
            )
            positivity_constraint = LinearConstraint(
                np.eye(bitstring_length).astype(int), lb=0.0, ub=1.0
            )
            constraints = [normality_constraint, positivity_constraint]
            # Minimize the corrected probabilities.
            res = minimize(
                corrected_probas, p_corr0, args=(T_matrix, p_raw), constraints=constraints
            )
            p_corr = res.x

        elif optimization_type == ReadOutOptimization.MLE:
            # Compute corrected inverse using matrix inversion and run MLE.
            p_corr = mle_solve(T_inv @ p_raw)
        else:
            raise NotImplementedError(
                f"Requested method {optimization_type} does not match supported protocols."
            )

        corrected_counts = np.rint(p_corr * n_shots).astype(int)

        # Renormalize if total counts differs from n_shots.
        corrected_counts = renormalize_counts(corrected_counts=corrected_counts, n_shots=n_shots)
        # At this point, the count should be off by at most 2, added or substracted to/from the
        # max count.
        if sum(corrected_counts) != n_shots:
            count_diff = sum(corrected_counts) - n_shots
            max_count_bs = np.argmax(corrected_counts)
            corrected_counts[max_count_bs] -= count_diff
        assert (
            corrected_counts.sum() == n_shots
        ), f"Corrected counts sum: {corrected_counts.sum()}, n_shots: {n_shots}"
        corrected_counters.append(
            Counter(
                {bs: count for bs, count in zip(ordered_bitstrings, corrected_counts) if count > 0}
            )
        )
    return corrected_counters


def mitigate(noise: Noise, mitigation: Mitigations, samples: list[Counter]) -> list[Counter]:
    return mitigation_minimization(noise=noise, mitigation=mitigation, samples=samples)



from __future__ import annotations

from collections import Counter
from typing import List

from torch import Tensor

from qadence.blocks import AbstractBlock
from qadence.blocks.utils import unroll_block_with_scaling
from qadence.measurements.utils import pauli_z_expectation


def compute_expectation(
    observable: List[AbstractBlock],
    samples: List[Counter],
) -> List[Tensor]:
    """Given a list of observables in Z basis, compute the expectation values against samples.

    Args:
        observables: A list of observables
            to estimate the expectation values from.
        samples: List of samples against which expectation value is to be computed
    """

    if not isinstance(observable, list):
        raise TypeError(
            f"Observables must be of type <class 'List[AbstractBlock]'>. Got {type(observable)}."
        )

    expectation_vals = []

    for obs in observable:
        decomposition = unroll_block_with_scaling(obs)
        if not obs._is_diag_pauli or obs.is_identity:
            raise TypeError("observable provided is not in the Z basis")
        expectation_vals.append(pauli_z_expectation(decomposition, samples))

    return expectation_vals



from __future__ import annotations

import importlib
from dataclasses import dataclass
from typing import Callable, cast

PROTOCOL_TO_MODULE = {
    "shadow": "qadence.measurements.shadow",
    "tomography": "qadence.measurements.tomography",
}


# TODO: make this a StrEnum to keep consistency with the rest of the interface
@dataclass
class Measurements:
    SHADOW = "shadow"
    TOMOGRAPHY = "tomography"

    def __init__(self, protocol: str, options: dict) -> None:
        self.protocol: str = protocol
        self.options: dict = options

    def get_measurement_fn(self) -> Callable:
        try:
            module = importlib.import_module(PROTOCOL_TO_MODULE[self.protocol])
        except KeyError:
            ImportError(f"The module corresponding to the protocol {self.protocol} is not found.")
        fn = getattr(module, "compute_expectation")
        return cast(Callable, fn)

    def _to_dict(self) -> dict:
        return {"protocol": self.protocol, "options": self.options}

    @classmethod
    def _from_dict(cls, d: dict) -> Measurements | None:
        if d:
            return cls(d["protocol"], **d["options"])
        return None



from __future__ import annotations

import torch
from torch import Tensor
from torch.nn import Module


def get_parameters(model: Module) -> Tensor:
    """Retrieve all trainable model parameters in a single vector.

    Args:
        model (Module): the input PyTorch model

    Returns:
        Tensor: a 1-dimensional tensor with the parameters
    """
    ps = [p.reshape(-1) for p in model.parameters() if p.requires_grad]
    return torch.concat(ps)


def set_parameters(model: Module, theta: Tensor) -> None:
    """Set all trainable parameters of a model from a single vector.

    Notice that this function assumes prior knowledge of right number
    of parameters in the model

    Args:
        model (Module): the input PyTorch model
        theta (Tensor): the parameters to assign
    """

    with torch.no_grad():
        idx = 0
        for ps in model.parameters():
            if ps.requires_grad:
                n = torch.numel(ps)
                if ps.ndim == 0:
                    ps[()] = theta[idx : idx + n]
                else:
                    ps[:] = theta[idx : idx + n].reshape(ps.size())
                idx += n


def num_parameters(model: Module) -> int:
    """Return the total number of parameters of the given model."""
    return len(get_parameters(model))



from __future__ import annotations

from typing import Callable, Iterable, List

import sympy
from numpy import array as nparray
from numpy import cdouble as npcdouble
from torch import tensor

from qadence.blocks import (
    AbstractBlock,
)
from qadence.blocks.utils import (
    expressions,
    parameters,
    uuid_to_expression,
)
from qadence.parameters import evaluate, make_differentiable, stringify
from qadence.types import ArrayLike, DifferentiableExpression, Engine, ParamDictType, TNumber


def _concretize_parameter(engine: Engine) -> Callable:
    if engine == Engine.JAX:
        from jax.numpy import array as jaxarray
        from jax.numpy import float64 as jaxfloat64

        def concretize_parameter(value: TNumber, trainable: bool = False) -> ArrayLike:
            return jaxarray([value], dtype=jaxfloat64)

    else:

        def concretize_parameter(value: TNumber, trainable: bool = False) -> ArrayLike:
            return tensor([value], requires_grad=trainable)

    return concretize_parameter


def unique(x: Iterable) -> List:
    return list(set(x))


def embedding(
    block: AbstractBlock, to_gate_params: bool = False, engine: Engine = Engine.TORCH
) -> tuple[ParamDictType, Callable[[ParamDictType, ParamDictType], ParamDictType],]:
    """Construct embedding function which maps user-facing parameters to either *expression-level*.

    parameters or *gate-level* parameters. The constructed embedding function has the signature:

         embedding_fn(params: ParamDictType, inputs: ParamDictType) -> ParamDictType:

    which means that it maps the *variational* parameter dict `params` and the *feature* parameter
    dict `inputs` to one new parameter dict `embedded_dict` which holds all parameters that are
    needed to execute a circuit on a given backend. There are two different *modes* for this
    mapping:

    - *Expression-level* parameters: For AD-based optimization. For every unique expression we end
      up with one entry in the embedded dict:
      `len(embedded_dict) == len(unique_parameter_expressions)`.
    - *Gate-level* parameters: For PSR-based optimization or real devices. One parameter for each
      gate parameter, regardless if they are based on the same expression. `len(embedded_dict) ==
      len(parametric_gates)`. This is needed because PSR requires to shift the angles of **every**
      gate where the same parameter appears.

    Arguments:
        block: parametrized block into which we want to embed parameters.
        to_gate_params: A boolean flag whether to generate gate-level parameters or
            expression-level parameters.

    Returns:
        A tuple with variational parameter dict and the embedding function.
    """
    concretize_parameter = _concretize_parameter(engine)
    if engine == Engine.TORCH:
        cast_dtype = tensor
    else:
        from jax.numpy import array

        cast_dtype = array

    unique_expressions = unique(expressions(block))
    unique_symbols = [p for p in unique(parameters(block)) if not isinstance(p, sympy.Array)]
    unique_const_matrices = [e for e in unique_expressions if isinstance(e, sympy.Array)]
    unique_expressions = [e for e in unique_expressions if not isinstance(e, sympy.Array)]

    # NOTE
    # there are 3 kinds of parameters in qadence
    # - non-trainable which are considered as inputs for classical data
    # - trainable which are the variational parameters to be optimized
    # - fixed: which are non-trainable parameters with fixed value (e.g. pi/2)
    #
    # both non-trainable and trainable parameters can have the same element applied
    # to different operations in the quantum circuit, e.g. assigning the same parameter
    # to multiple gates.
    non_numeric_symbols = [p for p in unique_symbols if not p.is_number]
    trainable_symbols = [p for p in non_numeric_symbols if p.trainable]
    constant_expressions = [expr for expr in unique_expressions if expr.is_number]
    # we dont need to care about constant symbols if they are contained in an symbolic expression
    # we only care about gate params which are ONLY a constant

    embeddings: dict[sympy.Expr, DifferentiableExpression] = {
        expr: make_differentiable(expr=expr, engine=engine)
        for expr in unique_expressions
        if not expr.is_number
    }

    uuid_to_expr = uuid_to_expression(block)

    def embedding_fn(params: ParamDictType, inputs: ParamDictType) -> ParamDictType:
        embedded_params: dict[sympy.Expr, ArrayLike] = {}
        for expr, fn in embeddings.items():
            angle: ArrayLike
            values = {}
            for symbol in expr.free_symbols:
                if not symbol.is_time:
                    if symbol.name in inputs:
                        value = inputs[symbol.name]
                    elif symbol.name in params:
                        value = params[symbol.name]
                    else:
                        msg_trainable = "Trainable" if symbol.trainable else "Non-trainable"
                        raise KeyError(
                            f"{msg_trainable} parameter '{symbol.name}' not found in the "
                            f"inputs list: {list(inputs.keys())} nor the "
                            f"params list: {list(params.keys())}."
                        )
                    values[symbol.name] = value
                else:
                    values[symbol.name] = tensor(1.0)
            angle = fn(**values)
            # do not reshape parameters which are multi-dimensional
            # tensors, such as for example generator matrices
            if not len(angle.squeeze().shape) > 1:
                angle = angle.reshape(-1)
            embedded_params[expr] = angle

        for e in constant_expressions + unique_const_matrices:
            embedded_params[e] = params[stringify(e)]

        if to_gate_params:
            gate_lvl_params: ParamDictType = {}
            for uuid, e in uuid_to_expr.items():
                gate_lvl_params[uuid] = embedded_params[e]
            return gate_lvl_params
        else:
            out = {stringify(k): v for k, v in embedded_params.items()}
            out.update({"orig_param_values": inputs})
            return out

    params: ParamDictType
    params = {
        p.name: concretize_parameter(value=p.value, trainable=True) for p in trainable_symbols
    }
    params.update(
        {
            stringify(expr): concretize_parameter(value=evaluate(expr), trainable=False)
            for expr in constant_expressions
        }
    )
    params.update(
        {
            stringify(expr): cast_dtype(nparray(expr.tolist(), dtype=npcdouble))
            for expr in unique_const_matrices
        }
    )
    return params, embedding_fn



from __future__ import annotations

import importlib
import math
from logging import getLogger
from typing import Callable, Union

from rich.progress import (
    BarColumn,
    Progress,
    TaskProgressColumn,
    TextColumn,
    TimeRemainingColumn,
)
from torch import complex128, float32, float64
from torch import device as torch_device
from torch import dtype as torch_dtype
from torch.nn import DataParallel, Module
from torch.optim import Optimizer
from torch.utils.data import DataLoader
from torch.utils.tensorboard import SummaryWriter

from qadence.ml_tools.config import TrainConfig
from qadence.ml_tools.data import DictDataLoader, data_to_device
from qadence.ml_tools.optimize_step import optimize_step
from qadence.ml_tools.printing import (
    log_model_tracker,
    log_tracker,
    plot_tracker,
    print_metrics,
    write_tracker,
)
from qadence.ml_tools.saveload import load_checkpoint, write_checkpoint
from qadence.types import ExperimentTrackingTool

logger = getLogger(__name__)


def train(
    model: Module,
    dataloader: Union[None, DataLoader, DictDataLoader],
    optimizer: Optimizer,
    config: TrainConfig,
    loss_fn: Callable,
    device: torch_device = None,
    optimize_step: Callable = optimize_step,
    dtype: torch_dtype = None,
) -> tuple[Module, Optimizer]:
    """Runs the training loop with gradient-based optimizer.

    Assumes that `loss_fn` returns a tuple of (loss,
    metrics: dict), where `metrics` is a dict of scalars. Loss and metrics are
    written to tensorboard. Checkpoints are written every
    `config.checkpoint_every` steps (and after the last training step).  If a
    checkpoint is found at `config.folder` we resume training from there.  The
    tensorboard logs can be viewed via `tensorboard --logdir /path/to/folder`.

    Args:
        model: The model to train.
        dataloader: dataloader of different types. If None, no data is required by
            the model
        optimizer: The optimizer to use.
        config: `TrainConfig` with additional training options.
        loss_fn: Loss function returning (loss: float, metrics: dict[str, float], ...)
        device: String defining device to train on, pass 'cuda' for GPU.
        optimize_step: Customizable optimization callback which is called at every iteration.=
            The function must have the signature `optimize_step(model,
            optimizer, loss_fn, xs, device="cpu")`.
        dtype: The dtype to use for the data.

    Example:
    ```python exec="on" source="material-block"
    from pathlib import Path
    import torch
    from itertools import count
    from qadence import Parameter, QuantumCircuit, Z
    from qadence import hamiltonian_factory, hea, feature_map, chain
    from qadence import QNN
    from qadence.ml_tools import TrainConfig, train_with_grad, to_dataloader

    n_qubits = 2
    fm = feature_map(n_qubits)
    ansatz = hea(n_qubits=n_qubits, depth=3)
    observable = hamiltonian_factory(n_qubits, detuning = Z)
    circuit = QuantumCircuit(n_qubits, fm, ansatz)

    model = QNN(circuit, observable, backend="pyqtorch", diff_mode="ad")
    batch_size = 1
    input_values = {"phi": torch.rand(batch_size, requires_grad=True)}
    pred = model(input_values)

    ## lets prepare the train routine

    cnt = count()
    criterion = torch.nn.MSELoss()
    optimizer = torch.optim.Adam(model.parameters(), lr=0.1)

    def loss_fn(model: torch.nn.Module, data: torch.Tensor) -> tuple[torch.Tensor, dict]:
        next(cnt)
        x, y = data[0], data[1]
        out = model(x)
        loss = criterion(out, y)
        return loss, {}

    tmp_path = Path("/tmp")
    n_epochs = 5
    batch_size = 25
    config = TrainConfig(
        folder=tmp_path,
        max_iter=n_epochs,
        checkpoint_every=100,
        write_every=100,
    )
    x = torch.linspace(0, 1, batch_size).reshape(-1, 1)
    y = torch.sin(x)
    data = to_dataloader(x, y, batch_size=batch_size, infinite=True)
    train_with_grad(model, data, optimizer, config, loss_fn=loss_fn)
    ```
    """
    # load available checkpoint
    init_iter = 0
    log_device = "cpu" if device is None else device
    if config.folder:
        model, optimizer, init_iter = load_checkpoint(
            config.folder, model, optimizer, device=log_device
        )
        logger.debug(f"Loaded model and optimizer from {config.folder}")

    # Move model to device before optimizer is loaded
    if isinstance(model, DataParallel):
        model = model.module.to(device=device, dtype=dtype)
    else:
        model = model.to(device=device, dtype=dtype)
    # initialize tracking tool
    if config.tracking_tool == ExperimentTrackingTool.TENSORBOARD:
        writer = SummaryWriter(config.folder, purge_step=init_iter)
    else:
        writer = importlib.import_module("mlflow")

    perform_val = isinstance(config.val_every, int)
    if perform_val:
        if not isinstance(dataloader, DictDataLoader):
            raise ValueError(
                "If `config.val_every` is provided as an integer, dataloader must"
                "be an instance of `DictDataLoader`."
            )
        iter_keys = dataloader.dataloaders.keys()
        if "train" not in iter_keys or "val" not in iter_keys:
            raise ValueError(
                "If `config.val_every` is provided as an integer, the dictdataloader"
                "must have `train` and `val` keys to access the respective dataloaders."
            )
        val_dataloader = dataloader.dataloaders["val"]
        dataloader = dataloader.dataloaders["train"]

    ## Training
    progress = Progress(
        TextColumn("[progress.description]{task.description}"),
        BarColumn(),
        TaskProgressColumn(),
        TimeRemainingColumn(elapsed_when_finished=True),
    )
    data_dtype = None
    if dtype:
        data_dtype = float64 if dtype == complex128 else float32

    best_val_loss = math.inf

    with progress:
        dl_iter = iter(dataloader) if dataloader is not None else None

        # Initial validation evaluation
        try:
            if perform_val:
                dl_iter_val = iter(val_dataloader) if val_dataloader is not None else None
                xs = next(dl_iter_val)
                xs_to_device = data_to_device(xs, device=device, dtype=data_dtype)
                best_val_loss, metrics = loss_fn(model, xs_to_device)

                metrics["val_loss"] = best_val_loss
                write_tracker(writer, None, metrics, init_iter, tracking_tool=config.tracking_tool)

            if config.folder:
                if config.checkpoint_best_only:
                    write_checkpoint(config.folder, model, optimizer, iteration="best")
                else:
                    write_checkpoint(config.folder, model, optimizer, init_iter)

            plot_tracker(
                writer,
                model,
                init_iter,
                config.plotting_functions,
                tracking_tool=config.tracking_tool,
            )

        except KeyboardInterrupt:
            logger.info("Terminating training gracefully after the current iteration.")

        # outer epoch loop
        init_iter += 1
        for iteration in progress.track(range(init_iter, init_iter + config.max_iter)):
            try:
                # in case there is not data needed by the model
                # this is the case, for example, of quantum models
                # which do not have classical input data (e.g. chemistry)
                if dataloader is None:
                    loss, metrics = optimize_step(
                        model=model,
                        optimizer=optimizer,
                        loss_fn=loss_fn,
                        xs=None,
                        device=device,
                        dtype=data_dtype,
                    )
                    loss = loss.item()

                elif isinstance(dataloader, (DictDataLoader, DataLoader)):
                    loss, metrics = optimize_step(
                        model=model,
                        optimizer=optimizer,
                        loss_fn=loss_fn,
                        xs=next(dl_iter),  # type: ignore[arg-type]
                        device=device,
                        dtype=data_dtype,
                    )

                else:
                    raise NotImplementedError(
                        f"Unsupported dataloader type: {type(dataloader)}. "
                        "You can use e.g. `qadence.ml_tools.to_dataloader` to build a dataloader."
                    )

                if iteration % config.print_every == 0 and config.verbose:
                    # Note that the loss returned by optimize_step
                    # is the value before doing the training step
                    # which is printed accordingly by the previous iteration number
                    print_metrics(loss, metrics, iteration - 1)

                if iteration % config.write_every == 0:
                    write_tracker(
                        writer, loss, metrics, iteration, tracking_tool=config.tracking_tool
                    )

                if iteration % config.plot_every == 0:
                    plot_tracker(
                        writer,
                        model,
                        iteration,
                        config.plotting_functions,
                        tracking_tool=config.tracking_tool,
                    )
                if perform_val:
                    if iteration % config.val_every == 0:
                        xs = next(dl_iter_val)
                        xs_to_device = data_to_device(xs, device=device, dtype=data_dtype)
                        val_loss, *_ = loss_fn(model, xs_to_device)
                        if config.validation_criterion(val_loss, best_val_loss, config.val_epsilon):  # type: ignore[misc]
                            best_val_loss = val_loss
                            if config.folder and config.checkpoint_best_only:
                                write_checkpoint(config.folder, model, optimizer, iteration="best")
                            metrics["val_loss"] = val_loss
                            write_tracker(
                                writer, loss, metrics, iteration, tracking_tool=config.tracking_tool
                            )

                if config.folder:
                    if iteration % config.checkpoint_every == 0 and not config.checkpoint_best_only:
                        write_checkpoint(config.folder, model, optimizer, iteration)

            except KeyboardInterrupt:
                logger.info("Terminating training gracefully after the current iteration.")
                break

        # Handling printing the last training loss
        # as optimize_step does not give the loss value at the last iteration
        try:
            xs = next(dl_iter) if dataloader is not None else None  # type: ignore[arg-type]
            xs_to_device = data_to_device(xs, device=device, dtype=data_dtype)
            loss, metrics, *_ = loss_fn(model, xs_to_device)
            if dataloader is None:
                loss = loss.item()
            if iteration % config.print_every == 0 and config.verbose:
                print_metrics(loss, metrics, iteration)

        except KeyboardInterrupt:
            logger.info("Terminating training gracefully after the current iteration.")

    # Final checkpointing and writing
    if config.folder and not config.checkpoint_best_only:
        write_checkpoint(config.folder, model, optimizer, iteration)
    write_tracker(writer, loss, metrics, iteration, tracking_tool=config.tracking_tool)

    # writing hyperparameters
    if config.hyperparams:
        log_tracker(writer, config.hyperparams, metrics, tracking_tool=config.tracking_tool)

    # logging the model
    if config.log_model:
        log_model_tracker(writer, model, dataloader, tracking_tool=config.tracking_tool)

    # close tracker
    if config.tracking_tool == ExperimentTrackingTool.TENSORBOARD:
        writer.close()
    elif config.tracking_tool == ExperimentTrackingTool.MLFLOW:
        writer.end_run()

    return model, optimizer



from __future__ import annotations

from logging import getLogger

import torch

logger = getLogger(__name__)


def _k_d(a: int, b: int) -> int:
    """Kronecker delta."""
    return int(a == b)


def _ix_map(n: int, a: int, b: int) -> int:
    """Maps `(a, b)` with `b` in [1, n] and `a < b` to range [1, n(n-1)/2]."""
    return int(n * (a - 1) - 0.5 * a * (a + 1) + b)


def _build_matrix_M(n_qubits: int) -> torch.Tensor:
    """Sign matrix used by the DAQC technique for the Ising model."""
    flat_size = int(0.5 * n_qubits * (n_qubits - 1))

    def matrix_M_ix(j: int, k: int, n: int, m: int) -> float:
        return (-1.0) ** (_k_d(n, j) + _k_d(n, k) + _k_d(m, j) + _k_d(m, k))

    M = torch.zeros(flat_size, flat_size)
    for k in range(2, n_qubits + 1):
        for j in range(1, k):
            for m in range(2, n_qubits + 1):
                for n in range(1, m):
                    alpha = _ix_map(n_qubits, n, m)
                    beta = _ix_map(n_qubits, j, k)
                    M[alpha - 1, beta - 1] = matrix_M_ix(j, k, n, m)
    return M



from __future__ import annotations

import logging
import logging.config
import os
from importlib import import_module
from pathlib import Path

import yaml
from torch import cdouble, set_default_dtype
from torch import float64 as torchfloat64

DEFAULT_FLOAT_DTYPE = torchfloat64
DEFAULT_COMPLEX_DTYPE = cdouble
set_default_dtype(DEFAULT_FLOAT_DTYPE)

logging_levels = {
    "DEBUG": logging.DEBUG,
    "INFO": logging.INFO,
    "WARNING": logging.WARNING,
    "ERROR": logging.ERROR,
    "CRITICAL": logging.CRITICAL,
}
LOG_CONFIG_PATH = os.environ.get("QADENCE_LOG_CONFIG", f"{Path(__file__).parent}/log_config.yaml")
LOG_BASE_LEVEL = os.environ.get("QADENCE_LOG_LEVEL", "").upper()

with open(LOG_CONFIG_PATH, "r") as stream:
    log_config = yaml.load(stream, Loader=yaml.FullLoader)
logging.config.dictConfig(log_config)

logger: logging.Logger = logging.getLogger(__name__)
LOG_LEVEL = logging_levels.get(LOG_BASE_LEVEL, logging.INFO)  # type: ignore[arg-type]
logger.setLevel(LOG_LEVEL)
[
    h.setLevel(LOG_LEVEL)  # type: ignore[func-returns-value]
    for h in logger.handlers
    if h.get_name() == "console"
]
logger.debug(f"Qadence logger successfully setup with log level {LOG_LEVEL}")

from .analog import *
from .backend import *
from .backends import *
from .blocks import *
from .circuit import *
from .constructors import *
from .engines import *
from .exceptions import *
from .execution import *
from .measurements import *
from .ml_tools import *
from .model import *
from .noise import *
from .operations import *
from .overlap import *
from .parameters import *
from .register import *
from .serialization import *
from .states import *
from .transpile import *
from .types import *
from .utils import *

"""Fetch the functions defined in the __all__ of each sub-module.

Import to the qadence name space. Make sure each added submodule has the respective definition:

    - `__all__ = ["function0", "function1", ...]`

Furthermore, add the submodule to the list below to automatically build
the __all__ of the qadence namespace. Make sure to keep alphabetical ordering.
"""

list_of_submodules = [
    ".analog",
    ".backends",
    ".blocks",
    ".circuit",
    ".constructors",
    ".noise",
    ".exceptions",
    ".execution",
    ".measurements",
    ".ml_tools",
    ".model",
    ".operations",
    ".overlap",
    ".parameters",
    ".register",
    ".serialization",
    ".states",
    ".transpile",
    ".types",
    ".utils",
]

__all__ = []
for submodule in list_of_submodules:
    __all_submodule__ = getattr(import_module(submodule, package="qadence"), "__all__")
    __all__ += __all_submodule__



from __future__ import annotations

import importlib
from enum import Enum
from typing import Callable, Iterable, Tuple, Union

import numpy as np
import sympy
from matplotlib.figure import Figure
from numpy.typing import ArrayLike
from pyqtorch.utils import SolverType
from torch import Tensor, pi
from torch.nn import Module

TNumber = Union[int, float, complex, np.int64, np.float64]
"""Union of python and numpy numeric types."""

TDrawColor = Tuple[float, float, float, float]

TParameter = Union[TNumber, Tensor, sympy.Basic, str]
"""Union of numbers, tensors, and parameter types."""

TArray = Union[Iterable, Tensor, np.ndarray]
"""Union of common array types."""

TGenerator = Union[Tensor, sympy.Array, sympy.Basic]
"""Union of torch tensors and numpy arrays."""

PI = pi

# Modules to be automatically added to the qadence namespace
__all__ = [
    "AnsatzType",
    "Endianness",
    "Strategy",
    "ResultType",
    "ParameterType",
    "BackendName",
    "StateGeneratorType",
    "LTSOrder",
    "MultivariateStrategy",
    "ReuploadScaling",
    "BasisSet",
    "TensorType",
    "DiffMode",
    "BackendName",
    "Interaction",
    "DeviceType",
    "OverlapMethod",
    "AlgoHEvo",
    "SerializationFormat",
    "PI",
    "SolverType",
]  # type: ignore


class StrEnum(str, Enum):
    def __str__(self) -> str:
        """Used when dumping enum fields in a schema."""
        ret: str = self.value
        return ret

    @classmethod
    def list(cls) -> list[str]:
        return list(map(lambda c: c.value, cls))  # type: ignore


class Strategy(StrEnum):
    """Computing paradigm."""

    DIGITAL = "Digital"
    """Use the digital paradigm."""
    ANALOG = "Analog"
    """Use the analog paradigm."""
    SDAQC = "sDAQC"
    """Use the step-wise digital-analog QC paradigm."""
    BDAQC = "bDAQC"
    """Use the banged digital-analog QC paradigm."""
    RYDBERG = "Rydberg"
    """Use the Rydberg QC paradigm."""


class Endianness(StrEnum):
    """The endianness convention to use."""

    BIG = "Big"
    """Use Big endianness."""
    LITTLE = "Little"
    """Use little endianness."""


class ResultType(StrEnum):
    """Available data types for generating certain results."""

    STRING = "String"
    """String Type."""
    TORCH = "Torch"
    """Torch Tensor Type."""
    NUMPY = "Numpy"
    """Numpy Array Type."""


class ParameterType(StrEnum):
    """Parameter types available in qadence."""

    FEATURE = "Feature"
    """FeatureParameters act as input and are not trainable."""
    VARIATIONAL = "Variational"
    """VariationalParameters are trainable."""
    FIXED = "Fixed"
    """Fixed/ constant parameters are neither trainable nor act as input."""


class TensorType(StrEnum):
    """Tensor Types for converting blocks to tensors."""

    SPARSEDIAGONAL = "SparseDiagonal"
    """Convert a diagonal observable block to a sparse diagonal if possible."""
    DENSE = "Dense"
    """Convert a block to a dense tensor."""
    SPARSE = "Sparse"
    """Convert a observable block to a sparse tensor."""


class LTSOrder(StrEnum):
    """Lie-Trotter-Suzuki approximation order."""

    BASIC = "BASIC"
    """Basic."""
    ST2 = "ST2"
    """ST2."""
    ST4 = "ST4"
    """ST4."""


class BasisSet(StrEnum):
    """Basis set for feature maps."""

    FOURIER = "Fourier"
    """Fourier basis set."""
    CHEBYSHEV = "Chebyshev"
    """Chebyshev polynomials of the first kind."""


class ReuploadScaling(StrEnum):
    """Scaling for data reuploads in feature maps."""

    CONSTANT = "Constant"
    """Constant scaling."""
    TOWER = "Tower"
    """Linearly increasing scaling."""
    EXP = "Exponential"
    """Exponentially increasing scaling."""


class MultivariateStrategy(StrEnum):
    """Multivariate strategy for feature maps."""

    PARALLEL = "parallel"
    """Parallel strategy."""
    SERIES = "SERIES"
    """Serial strategy."""


class AnsatzType(StrEnum):
    """Ansatz types for variational circuits."""

    HEA = "hea"
    """Hardware-efficient ansatz."""
    IIA = "iia"
    """Identity-Initialised Ansatz."""


class _DiffMode(StrEnum):
    """Differentiation modes to choose from."""

    GPSR = "gpsr"
    """Basic generalized parameter shift rule."""
    AD = "ad"
    """Automatic Differentiation."""
    ADJOINT = "adjoint"
    """Adjoint Differentiation."""


class QubitSupportType(StrEnum):
    """Qubit support types."""

    GLOBAL = "global"
    """Use global qubit support."""


class Interaction(StrEnum):
    """Interaction types used in.

    - `RydbergDevice`.
    - [`hamiltonian_factory`][qadence.constructors.hamiltonians.hamiltonian_factory].
    """

    ZZ = "ZZ"
    """ZZ-Ising Interaction."""
    NN = "NN"
    """NN-Ising Interaction, N=(I-Z)/2."""
    XY = "XY"
    """XY Interaction."""
    XYZ = "XYZ"
    """XYZ Interaction."""


class DeviceType(StrEnum):
    """Supported types of devices for Pulser backend."""

    IDEALIZED = "IdealDevice"
    """Idealized device, least realistic."""

    REALISTIC = "RealisticDevice"
    """Device with realistic specs."""


class _BackendName(StrEnum):
    """The available backends for running circuits."""

    PYQTORCH = "pyqtorch"
    """The Pyqtorch backend."""
    BRAKET = "braket"
    """The Braket backend."""
    PULSER = "pulser"
    """The Pulser backend."""
    HORQRUX = "horqrux"
    """The horqrux backend."""


class _Engine(StrEnum):
    TORCH = "torch"
    JAX = "jax"


# If proprietary qadence_extensions is available, import the
# right function since more backends are supported.
try:
    module = importlib.import_module("qadence_extensions.types")
    BackendName = getattr(module, "BackendName")
    DiffMode = getattr(module, "DiffMode")
    Engine = getattr(module, "Engine")
except ModuleNotFoundError:
    BackendName = _BackendName
    DiffMode = _DiffMode
    Engine = _Engine


class StateGeneratorType(StrEnum):
    """Methods to generate random states."""

    RANDOM_ROTATIONS = "RandomRotations"
    """Random Rotations."""
    HAAR_MEASURE_FAST = "HaarMeasureFast"
    """HaarMeasure."""
    HAAR_MEASURE_SLOW = "HaarMeasureSlow"
    """HaarMeasure non-optimized version."""


class SerializationFormat(StrEnum):
    """Available serialization formats for circuits."""

    PT = "PT"
    """The PT format used by Torch."""
    JSON = "JSON"
    """The Json format."""


class OverlapMethod(StrEnum):
    """Overlap Methods to choose from."""

    EXACT = "exact"
    """Exact."""
    JENSEN_SHANNON = "jensen_shannon"
    """Jensen-shannon."""
    COMPUTE_UNCOMPUTE = "compute_uncompute"
    """Compute-uncompute."""
    SWAP_TEST = "swap_test"
    """Swap-test."""
    HADAMARD_TEST = "hadamard_test"
    """Hadamard-test."""


class FigFormat(StrEnum):
    """Available output formats for exporting visualized circuits to a file."""

    PNG = "PNG"
    """PNG format."""
    PDF = "PDF"
    """PDF format."""
    SVG = "SVG"
    """SVG format."""


class AlgoHEvo(StrEnum):
    """Hamiltonian Evolution algorithms that can be used by the backend."""

    RK4 = "RK4"
    """4th order Runge-Kutta approximation."""
    EIG = "EIG"
    """Using Hamiltonian diagonalization."""
    EXP = "EXP"
    """Using torch.matrix_exp on the generator matrix."""


class LatticeTopology(StrEnum):
    """Lattice topologies to choose from for the register."""

    LINE = "line"
    """Line-format lattice."""
    SQUARE = "square"
    """Square lattice."""
    CIRCLE = "circle"
    """Circular lattice."""
    ALL_TO_ALL = "all_to_all"
    """All to all- connected lattice."""
    RECTANGULAR_LATTICE = "rectangular_lattice"
    """Rectangular-shaped lattice."""
    TRIANGULAR_LATTICE = "triangular_lattice"
    """Triangular-shaped shape."""
    HONEYCOMB_LATTICE = "honeycomb_lattice"
    """Honeycomb-shaped lattice."""
    ARBITRARY = "arbitrary"
    """Arbitrarily-shaped lattice."""


class GenDAQC(StrEnum):
    """The type of interaction for the DAQC transform."""

    ZZ = "ZZ"
    """ZZ"""
    NN = "NN"
    """NN"""


class OpName(StrEnum):
    """A list of all available of digital-analog operations."""

    # Digital operations
    X = "X"
    """The X gate."""
    Y = "Y"
    """The Y gate."""
    Z = "Z"
    """The Z gate."""
    N = "N"
    """The N = (1/2)(I-Z) operator."""
    H = "H"
    """The Hadamard gate."""
    I = "I"  # noqa
    """The Identity gate."""
    ZERO = "Zero"
    """The zero gate."""
    RX = "RX"
    """The RX gate."""
    RY = "RY"
    """The RY gate."""
    RZ = "RZ"
    """The RZ gate."""
    U = "U"
    """The U gate."""
    CNOT = "CNOT"
    """The CNOT gate."""
    CZ = "CZ"
    """The CZ gate."""
    MCZ = "MCZ"
    """The Multicontrol CZ gate."""
    HAMEVO = "HamEvo"
    """The Hamiltonian Evolution operation."""
    CRX = "CRX"
    """The Control RX gate."""
    MCRX = "MCRX"
    """The Multicontrol RX gate."""
    CRY = "CRY"
    """The Controlled RY gate."""
    MCRY = "MCRY"
    """The Multicontrol RY gate."""
    CRZ = "CRZ"
    """The Control RZ gate."""
    MCRZ = "MCRZ"
    """The Multicontrol RZ gate."""
    CSWAP = "CSWAP"
    """The Control SWAP gate."""
    T = "T"
    """The T gate."""
    # FIXME: Tdagger is not currently supported by any backend
    TDAGGER = "TDagger"
    """The T dagger gate."""
    S = "S"
    """The S gate."""
    SDAGGER = "SDagger"
    """The S dagger gate."""
    SWAP = "SWAP"
    """The SWAP gate."""
    PHASE = "PHASE"
    """The PHASE gate."""
    CPHASE = "CPHASE"
    """The controlled PHASE gate."""
    MCPHASE = "MCPHASE"
    """The Multicontrol PHASE gate."""
    TOFFOLI = "Toffoli"
    """The Toffoli gate."""
    # Analog operations
    ANALOGENTANG = "AnalogEntanglement"
    """The analog entanglement operation."""
    ANALOGRX = "AnalogRX"
    """The analog RX operation."""
    ANALOGRY = "AnalogRY"
    """The analog RY operation."""
    ANALOGRZ = "AnalogRZ"
    """The analog RZ operation."""
    ANALOGSWAP = "AnalogSWAP"
    """The analog SWAP operation."""
    ENTANGLE = "entangle"
    """The entanglement operation."""
    ANALOGINTERACTION = "AnalogInteraction"
    """The analog interaction operation."""
    PROJ = "Projector"
    """The projector operation."""


class ReadOutOptimization(StrEnum):
    MLE = "mle"
    CONSTRAINED = "constrained"


ParamDictType = dict[str, ArrayLike]
DifferentiableExpression = Callable[..., ArrayLike]


class InputDiffMode(StrEnum):
    """Derivative modes w.r.t inputs of UFAs."""

    AD = "ad"
    """Reverse automatic differentiation."""
    FD = "fd"
    """Central finite differencing."""


class ObservableTransform:
    """Observable transformation type."""

    SCALE = "scale"
    """Use the given values as scale and shift."""
    RANGE = "range"
    """Use the given values as min and max."""
    NONE = "none"
    """No transformation."""


class ExperimentTrackingTool(StrEnum):
    TENSORBOARD = "tensorboard"
    """Use the tensorboard experiment tracker."""
    MLFLOW = "mlflow"
    """Use the ml-flow experiment tracker."""


LoggablePlotFunction = Callable[[Module, int], tuple[str, Figure]]



from __future__ import annotations

from logging import getLogger
from typing import Any, Callable, Union

from matplotlib.figure import Figure
from mlflow.models import infer_signature
from torch import Tensor
from torch.nn import Module
from torch.utils.data import DataLoader
from torch.utils.tensorboard import SummaryWriter

from qadence.ml_tools.data import DictDataLoader
from qadence.types import ExperimentTrackingTool

logger = getLogger(__name__)

PlottingFunction = Callable[[Module, int], tuple[str, Figure]]
InputData = Union[Tensor, dict[str, Tensor]]


def print_metrics(loss: float | None, metrics: dict, iteration: int) -> None:
    msg = " ".join(
        [f"Iteration {iteration: >7} | Loss: {loss:.7f} -"]
        + [f"{k}: {v.item():.7f}" for k, v in metrics.items()]
    )
    print(msg)


def write_tensorboard(
    writer: SummaryWriter, loss: float = None, metrics: dict = {}, iteration: int = 0
) -> None:
    if loss is not None:
        writer.add_scalar("loss", loss, iteration)
    for key, arg in metrics.items():
        writer.add_scalar(key, arg, iteration)


def log_hyperparams_tensorboard(writer: SummaryWriter, hyperparams: dict, metrics: dict) -> None:
    writer.add_hparams(hyperparams, metrics)


def plot_tensorboard(
    writer: SummaryWriter,
    model: Module,
    iteration: int,
    plotting_functions: tuple[PlottingFunction],
) -> None:
    for pf in plotting_functions:
        descr, fig = pf(model, iteration)
        writer.add_figure(descr, fig, global_step=iteration)


def log_model_tensorboard(
    writer: SummaryWriter,
    model: Module,
    dataloader: Union[None, DataLoader, DictDataLoader],
) -> None:
    logger.warning("Model logging is not supported by tensorboard. No model will be logged.")


def write_mlflow(writer: Any, loss: float | None, metrics: dict, iteration: int) -> None:
    writer.log_metrics({"loss": float(loss)}, step=iteration)  # type: ignore
    writer.log_metrics(metrics, step=iteration)  # logs the single metrics


def log_hyperparams_mlflow(writer: Any, hyperparams: dict, metrics: dict) -> None:
    writer.log_params(hyperparams)  # type: ignore


def plot_mlflow(
    writer: Any,
    model: Module,
    iteration: int,
    plotting_functions: tuple[PlottingFunction],
) -> None:
    for pf in plotting_functions:
        descr, fig = pf(model, iteration)
        writer.log_figure(fig, descr)


def log_model_mlflow(
    writer: Any, model: Module, dataloader: DataLoader | DictDataLoader | None
) -> None:
    if dataloader is not None:
        xs: InputData
        xs, *_ = next(iter(dataloader))
        preds = model(xs)
        if isinstance(xs, Tensor):
            xs = xs.numpy()
            preds = preds.detach().numpy()
        elif isinstance(xs, dict):
            for key, val in xs.items():
                xs[key] = val.numpy()
            for key, val in preds.items():
                preds[key] = val.detach.numpy()
        signature = infer_signature(xs, preds)
    else:
        signature = None
    writer.pytorch.log_model(model, artifact_path="model", signature=signature)


TRACKER_MAPPING: dict[ExperimentTrackingTool, Callable[..., None]] = {
    ExperimentTrackingTool.TENSORBOARD: write_tensorboard,
    ExperimentTrackingTool.MLFLOW: write_mlflow,
}

LOGGER_MAPPING: dict[ExperimentTrackingTool, Callable[..., None]] = {
    ExperimentTrackingTool.TENSORBOARD: log_hyperparams_tensorboard,
    ExperimentTrackingTool.MLFLOW: log_hyperparams_mlflow,
}

PLOTTER_MAPPING: dict[ExperimentTrackingTool, Callable[..., None]] = {
    ExperimentTrackingTool.TENSORBOARD: plot_tensorboard,
    ExperimentTrackingTool.MLFLOW: plot_mlflow,
}

MODEL_LOGGER_MAPPING: dict[ExperimentTrackingTool, Callable[..., None]] = {
    ExperimentTrackingTool.TENSORBOARD: log_model_tensorboard,
    ExperimentTrackingTool.MLFLOW: log_model_mlflow,
}


def write_tracker(
    *args: Any, tracking_tool: ExperimentTrackingTool = ExperimentTrackingTool.TENSORBOARD
) -> None:
    return TRACKER_MAPPING[tracking_tool](*args)


def log_tracker(
    *args: Any, tracking_tool: ExperimentTrackingTool = ExperimentTrackingTool.TENSORBOARD
) -> None:
    return LOGGER_MAPPING[tracking_tool](*args)


def plot_tracker(
    *args: Any, tracking_tool: ExperimentTrackingTool = ExperimentTrackingTool.TENSORBOARD
) -> None:
    return PLOTTER_MAPPING[tracking_tool](*args)


def log_model_tracker(
    *args: Any, tracking_tool: ExperimentTrackingTool = ExperimentTrackingTool.TENSORBOARD
) -> None:
    return MODEL_LOGGER_MAPPING[tracking_tool](*args)



from __future__ import annotations

from .analog import (
    AnalogEntanglement,
    AnalogInteraction,
    AnalogRot,
    AnalogRX,
    AnalogRY,
    AnalogRZ,
    AnalogSWAP,
    ConstantAnalogRotation,
    entangle,
)
from .control_ops import (
    CNOT,
    CPHASE,
    CRX,
    CRY,
    CRZ,
    CSWAP,
    CZ,
    MCPHASE,
    MCRX,
    MCRY,
    MCRZ,
    MCZ,
    Toffoli,
)
from .ham_evo import HamEvo
from .parametric import PHASE, RX, RY, RZ, U
from .primitive import SWAP, H, I, N, Projector, S, SDagger, T, TDagger, X, Y, Z, Zero

# gate sets
# FIXME: this could be inferred by the number of qubits if we had
# a class property for each operation. The number of qubits can default
# to None for operations which do not have it by default
# this would allow to greatly simplify the tests
pauli_gateset: list = [I, X, Y, Z]
# FIXME: add Tdagger when implemented
single_qubit_gateset = [X, Y, Z, H, I, RX, RY, RZ, U, S, SDagger, T, PHASE]
two_qubit_gateset = [CNOT, SWAP, CZ, CRX, CRY, CRZ, CPHASE]
three_qubit_gateset = [CSWAP]
multi_qubit_gateset = [Toffoli, MCRX, MCRY, MCRZ, MCPHASE, MCZ]
analog_gateset = [
    HamEvo,
    ConstantAnalogRotation,
    AnalogEntanglement,
    AnalogSWAP,
    AnalogRX,
    AnalogRY,
    AnalogRZ,
    AnalogInteraction,
    entangle,
]
non_unitary_gateset = [Zero, N, Projector]


# Modules to be automatically added to the qadence namespace
__all__ = [
    "X",
    "Y",
    "Z",
    "N",
    "H",
    "I",
    "Zero",
    "RX",
    "RY",
    "RZ",
    "U",
    "CNOT",
    "CZ",
    "MCZ",
    "HamEvo",
    "CRX",
    "MCRX",
    "CRY",
    "MCRY",
    "CRZ",
    "MCRZ",
    "T",
    "TDagger",
    "S",
    "SDagger",
    "SWAP",
    "PHASE",
    "CPHASE",
    "CSWAP",
    "MCPHASE",
    "Toffoli",
    "entangle",
    "AnalogEntanglement",
    "AnalogInteraction",
    "AnalogRot",
    "AnalogRX",
    "AnalogRY",
    "AnalogRZ",
    "AnalogSWAP",
    "Projector",
]



from __future__ import annotations

from copy import deepcopy
from functools import singledispatch
from logging import getLogger
from typing import Callable, Iterable, Type

import sympy

from qadence.blocks import (
    AbstractBlock,
    AddBlock,
    AnalogBlock,
    ChainBlock,
    CompositeBlock,
    ControlBlock,
    KronBlock,
    PrimitiveBlock,
    PutBlock,
    ScaleBlock,
    add,
    chain,
    kron,
)
from qadence.blocks.utils import (
    TPrimitiveBlock,
    _construct,
    parameters,
)
from qadence.operations import SWAP, I
from qadence.parameters import Parameter

logger = getLogger(__name__)


def repeat(
    Block: Type[TPrimitiveBlock], support: Iterable[int], parameter: str | Parameter | None = None
) -> KronBlock:
    if parameter is None:
        return kron(Block(i) for i in support)  # type: ignore [arg-type]
    return kron(Block(i, parameter) for i in support)  # type: ignore [call-arg, arg-type]


def set_trainable(
    blocks: AbstractBlock | list[AbstractBlock], value: bool = True, inplace: bool = True
) -> AbstractBlock | list[AbstractBlock]:
    """Set the trainability of all parameters in a block to a given value.

    Args:
        blocks (AbstractBlock | list[AbstractBlock]): Block or list of blocks for which
            to set the trainable attribute
        value (bool, optional): The value of the trainable attribute to assign to the input blocks
        inplace (bool, optional): Whether to modify the block(s) in place or not. Currently, only

    Raises:
        NotImplementedError: if the `inplace` argument is set to False, the function will
            raise  this exception

    Returns:
        AbstractBlock | list[AbstractBlock]: the input block or list of blocks with the trainable
            attribute set to the given value
    """

    if isinstance(blocks, AbstractBlock):
        blocks = [blocks]

    if inplace:
        for block in blocks:
            params: list[sympy.Basic] = parameters(block)
            for p in params:
                if not p.is_number:
                    p.trainable = value
    else:
        raise NotImplementedError("Not inplace set_trainable is not yet available")

    return blocks if len(blocks) > 1 else blocks[0]


def validate(block: AbstractBlock) -> AbstractBlock:
    """Moves a block from global to local qubit numbers by adding PutBlocks.

    Reassigns qubit locations appropriately.

    # Example
    ```python exec="on" source="above" result="json"
    from qadence.blocks import chain
    from qadence.operations import X
    from qadence.transpile import validate

    x = chain(chain(X(0)), chain(X(1)))
    print(x)
    print(validate(x))
    ```
    """
    vblock: AbstractBlock
    from qadence.transpile import reassign

    if isinstance(block, ControlBlock):
        vblock = deepcopy(block)
        b: AbstractBlock
        (b,) = block.blocks
        b = reassign(b, {i: i - min(b.qubit_support) for i in b.qubit_support})
        b = validate(b)
        vblock.blocks = (b,)  # type: ignore[assignment]

    elif isinstance(block, CompositeBlock):
        blocks = []
        for b in block.blocks:
            mi, ma = min(b.qubit_support), max(b.qubit_support)
            nb = reassign(b, {i: i - min(b.qubit_support) for i in b.qubit_support})
            nb = validate(nb)
            nb = PutBlock(nb, tuple(range(mi, ma + 1)))
            blocks.append(nb)
        try:
            vblock = _construct(type(block), tuple(blocks))
        except AssertionError as e:
            if str(e) == "Make sure blocks act on distinct qubits!":
                vblock = chain(*blocks)
            else:
                raise e

    elif isinstance(block, PrimitiveBlock):
        vblock = deepcopy(block)

    else:
        raise NotImplementedError

    vblock.tag = block.tag
    return vblock


@singledispatch
def scale_primitive_blocks_only(block: AbstractBlock, scale: sympy.Basic = None) -> AbstractBlock:
    """Push the scale all the way down into the leaves of the block tree.

    When given a scaled CompositeBlock consisting of several PrimitiveBlocks.

    Arguments:
        block: The block to be transpiled.
        scale: An optional scale parameter. Only to be used for recursive calls internally.

    Returns:
        AbstractBlock: A block of the same type where the scales have been moved into the subblocks.

    Examples:

    There are two different cases:
    `ChainBlock`s/`KronBlock`s: Only the first subblock needs to be scaled because chains/krons
    represent multiplications.
    ```python exec="on" source="above" result="json"
    from qadence import chain, X, RX
    from qadence.transpile import scale_primitive_blocks_only
    b = 2 * chain(X(0), RX(0, "theta"))
    print(b)
    # After applying scale_primitive_blocks_only
    print(scale_primitive_blocks_only(b))
    ```

    `AddBlock`s: Consider 2 * add(X(0), RX(0, "theta")).  The scale needs to be added to all
    subblocks.  We get add(2 * X(0), 2 * RX(0, "theta")).
    ```python exec="on" source="above" result="json"
    from qadence import add, X, RX
    from qadence.transpile import scale_primitive_blocks_only
    b = 2 * add(X(0), RX(0, "theta"))
    print(b)
    # After applying scale_primitive_blocks_only
    print(scale_primitive_blocks_only(b))
    ```
    """
    raise NotImplementedError(f"scale_primitive_blocks_only is not implemented for {type(block)}")


@scale_primitive_blocks_only.register
def _(block: ScaleBlock, scale: sympy.Basic = None) -> AbstractBlock:
    (scale2,) = block.parameters.expressions()
    s = scale2 if scale is None else scale * scale2
    blk = scale_primitive_blocks_only(block.block, s)
    blk.tag = block.tag
    return blk


@scale_primitive_blocks_only.register
def _(block: ChainBlock, scale: sympy.Basic = None) -> CompositeBlock:
    blk = scale_only_first_block(chain, block, scale)
    blk.tag = block.tag
    return blk


@scale_primitive_blocks_only.register
def _(block: KronBlock, scale: sympy.Basic = None) -> CompositeBlock:
    blk = scale_only_first_block(kron, block, scale)
    blk.tag = block.tag
    return blk


@scale_primitive_blocks_only.register
def _(block: AddBlock, scale: sympy.Basic = None) -> CompositeBlock:
    blk = add(scale_primitive_blocks_only(b, scale) for b in block.blocks)
    blk.tag = block.tag
    return blk


@scale_primitive_blocks_only.register
def _(block: PrimitiveBlock, scale: sympy.Basic = None) -> AbstractBlock:
    if scale is None:
        return block
    b: ScaleBlock = block * scale
    return b


@scale_primitive_blocks_only.register
def _(block: AnalogBlock, scale: sympy.Basic = None) -> AbstractBlock:
    if scale is not None:
        raise NotImplementedError("Cannot scale `AnalogBlock`s!")
    return block


def scale_only_first_block(
    fn: Callable, block: CompositeBlock, scale: sympy.Basic = None
) -> CompositeBlock:
    if len(block.blocks):
        first, rest = block.blocks[0], block.blocks[1:]
        firstscaled = scale_primitive_blocks_only(first, scale)

        blk: CompositeBlock
        blk = fn(firstscaled, *[scale_primitive_blocks_only(b, None) for b in rest])
        return blk
    else:
        return block


@singledispatch
def fill_identities(block: AbstractBlock, start: int, stop: int) -> AbstractBlock:
    return block


@fill_identities.register
def _(block: PrimitiveBlock, start: int, stop: int) -> AbstractBlock:
    full_support = tuple(range(start, stop + 1))
    if block.qubit_support == full_support:
        return block
    tag = block.tag
    block.tag = None
    bs = [block] + [I(i) for i in (set(full_support) - set(block.qubit_support))]
    b = kron(*sorted(bs, key=lambda x: x.qubit_support))
    b.tag = tag
    return b


@fill_identities.register
def _(block: SWAP, start: int, stop: int) -> AbstractBlock:
    full_support = tuple(range(start, stop + 1))
    if block.qubit_support == full_support:
        return block
    tag = block.tag
    block.tag = None
    bs = [block] + [chain(I(i), I(i)) for i in (set(full_support) - set(block.qubit_support))]
    b = kron(*sorted(bs, key=lambda x: x.qubit_support))
    b.tag = tag
    return b


@fill_identities.register
def _(block: ChainBlock, start: int, stop: int) -> AbstractBlock:
    b = chain(fill_identities(b, start, stop) for b in block)
    b.tag = block.tag
    return b


def _fill_kron(block: KronBlock, start: int, stop: int) -> list[AbstractBlock]:
    def length(b: AbstractBlock | list) -> int:
        if isinstance(b, list):
            return max(map(length, b))
        elif isinstance(b, ChainBlock):
            return len(b)
        elif isinstance(b, SWAP):
            return 2
        else:
            return 1

    def append_ids(block: AbstractBlock, total: int) -> AbstractBlock:
        qs = block.qubit_support
        ids = [I(i) for i in range(min(qs), max(qs)) for _ in range(length(block), total)]
        bs = [block] + ids
        return chain(*bs)

    def id_chain(i: int, max_len: int) -> AbstractBlock:
        return chain(I(i) for _ in range(max_len))

    # fill subblocks with identities
    bs = [fill_identities(b, min(b.qubit_support), max(b.qubit_support)) for b in block]

    # make all subblocks equally "long" horizontally
    max_len = length(bs)
    bs = [append_ids(b, max_len) for b in bs]

    # fill potentially completely empty wires
    bs += [id_chain(i, max_len) for i in (set(range(start, stop + 1)) - set(block.qubit_support))]

    return sorted(bs, key=lambda x: x.qubit_support)


@fill_identities.register
def _(block: KronBlock, start: int, stop: int) -> AbstractBlock:
    b = kron(*_fill_kron(block, start, stop))  # type: ignore[misc]
    b.tag = block.tag
    return b


@fill_identities.register
def _(block: AddBlock, start: int, stop: int) -> AbstractBlock:
    b = add(fill_identities(b, start, stop) for b in block)
    b.tag = block.tag
    return b


def is_kron_of_primitives(block: AbstractBlock) -> bool:
    return isinstance(block, (KronBlock)) and all(
        [isinstance(b, PrimitiveBlock) and b.n_supports == 1 for b in block.blocks]
    )


def is_chain_of_primitivekrons(block: AbstractBlock) -> bool:
    if isinstance(block, ChainBlock) and len(block) > 1:
        same_length = len(set(map(len, block))) == 1
        kron_of_primitives_same_qubit = all(
            is_kron_of_primitives(b) and b.qubit_support == block.qubit_support
            for b in block.blocks
        )
        return same_length and kron_of_primitives_same_qubit
    else:
        return False


def chain_single_qubit_ops(block: AbstractBlock) -> AbstractBlock:
    """Transpile a chain of krons into a kron of chains of single qubit operations.

    Examples:
    ```python exec="on" source="above" result="json"
    from qadence import hea
    from qadence.transpile.block import chain_single_qubit_ops

    # Consider a single HEA layer
    block = hea(2,1)
    print(block)

    # After applying chain_single_qubit_ops, we get:
    print(chain_single_qubit_ops(block))
    ```
    """
    if is_chain_of_primitivekrons(block):
        try:
            return kron(*map(lambda bs: chain(*bs), zip(*block)))  # type: ignore[misc]
        except Exception as e:
            logger.debug(
                f"Unable to transpile {block} using chain_single_qubit_ops\
                         due to {e}. Returning original circuit."
            )
            return block

    elif isinstance(block, CompositeBlock):
        return _construct(type(block), tuple(chain_single_qubit_ops(b) for b in block.blocks))
    else:
        return block



from __future__ import annotations

from qadence.backend import Backend, BackendConfiguration
from qadence.engines.differentiable_backend import DifferentiableBackend
from qadence.extensions import (
    import_backend,
    import_config,
    import_engine,
    set_backend_config,
)
from qadence.logger import get_logger
from qadence.types import BackendName, DiffMode

__all__ = ["backend_factory", "config_factory"]
logger = get_logger(__name__)


def backend_factory(
    backend: BackendName | str,
    diff_mode: DiffMode | str | None = None,
    configuration: BackendConfiguration | dict | None = None,
) -> Backend | DifferentiableBackend:
    backend_inst: Backend | DifferentiableBackend
    try:
        BackendCls = import_backend(backend)
        default_config = BackendCls.default_configuration()
        if configuration is None:
            configuration = default_config
        elif isinstance(configuration, dict):
            configuration = config_factory(backend, configuration)
        else:
            # NOTE: types have to match exactly, hence we use `type`
            if not isinstance(configuration, type(BackendCls.default_configuration())):
                expected_cfg = BackendCls.default_configuration()
                raise ValueError(
                    f"Given config class '{type(configuration)}' does not match the backend",
                    f" class: '{BackendCls}'. Expected: '{type(expected_cfg)}.'",
                )

        # Instantiate the backend
        backend_inst = BackendCls(  # type: ignore[operator]
            config=configuration
            if configuration is not None
            else BackendCls.default_configuration()
        )
        set_backend_config(backend_inst, diff_mode)
        # Wrap the quantum Backend in a DifferentiableBackend if a diff_mode is passed.
        if diff_mode is not None:
            diff_backend_cls = import_engine(backend_inst.engine)
            backend_inst = diff_backend_cls(backend=backend_inst, diff_mode=DiffMode(diff_mode))  # type: ignore[operator]
        return backend_inst
    except Exception as e:
        msg = f"The requested backend '{backend}' is either not installed\
            or could not be imported due to {e}."
        logger.error(msg)
        raise Exception(msg)
    # Set backend configurations which depend on the differentiation mode


def config_factory(backend_name: BackendName | str, config: dict) -> BackendConfiguration:
    cfg: BackendConfiguration
    try:
        BackendConfigCls = import_config(backend_name)
        cfg = BackendConfigCls(**config)  # type: ignore[operator]
    except Exception as e:
        logger.debug(f"Unable to import config for backend {backend_name} due to {e}.")
    return cfg



from __future__ import annotations

from functools import partial
from logging import getLogger
from typing import Union

import numpy as np
from pulser.channels.base_channel import Channel
from pulser.parametrized.variable import Variable, VariableItem
from pulser.pulse import Pulse
from pulser.sequence.sequence import Sequence
from pulser.waveforms import CompositeWaveform, ConstantWaveform, RampWaveform

from qadence import Register
from qadence.analog import AddressingPattern
from qadence.blocks import AbstractBlock, CompositeBlock
from qadence.blocks.analog import (
    AnalogBlock,
    AnalogComposite,
    ConstantAnalogRotation,
    Interaction,
    InteractionBlock,
)
from qadence.operations import RX, RY, RZ, AnalogEntanglement
from qadence.parameters import evaluate
from qadence.types import PI, OpName

from .channels import GLOBAL_CHANNEL, LOCAL_CHANNEL
from .config import Configuration
from .waveforms import SquareWaveform

logger = getLogger(__name__)

TVar = Union[Variable, VariableItem]

supported_gates = [
    OpName.ZERO,
    OpName.RX,
    OpName.RY,
    OpName.RZ,
    OpName.ANALOGENTANG,
    OpName.ANALOGRX,
    OpName.ANALOGRY,
    OpName.ANALOGRZ,
    OpName.ANALOGSWAP,
    OpName.ANALOGINTERACTION,
]


def add_addressing_pattern(
    sequence: Sequence,
    pattern: AddressingPattern,
) -> None:
    total_duration = sequence.get_duration()
    n_qubits = len(sequence.register.qubits)

    support = tuple(range(n_qubits))

    amp = pattern.amp
    det = pattern.det
    weights_amp = pattern.weights_amp
    weights_det = pattern.weights_det
    local_constr_amp = pattern.local_constr_amp
    local_constr_det = pattern.local_constr_det
    global_constr_amp = pattern.global_constr_amp
    global_constr_det = pattern.global_constr_det

    for i in support:
        # declare separate local channel for each qubit
        sequence.declare_channel(f"ch_q{i}", "rydberg_local", initial_target=0)

    # add amplitude and detuning patterns
    for i in support:
        if weights_amp[i].is_number:  # type: ignore [union-attr]
            w_amp = evaluate(weights_amp[i], as_torch=True) * local_constr_amp[i]
        else:
            raise ValueError(
                "Pulser backend currently doesn't support parametrized amplitude pattern weights."
            )

        if weights_det[i].is_number:  # type: ignore [union-attr]
            w_det = evaluate(weights_det[i], as_torch=True) * local_constr_det[i]
        else:
            raise ValueError(
                "Pulser backend currently doesn't support parametrized detuning pattern weights."
            )

        omega = global_constr_amp * amp * w_amp
        detuning = global_constr_det * det * w_det
        pulse = Pulse.ConstantPulse(
            duration=total_duration, amplitude=omega, detuning=detuning, phase=0
        )
        sequence.target(i, f"ch_q{i}")
        sequence.add(pulse, f"ch_q{i}", protocol="no-delay")


def add_pulses(
    sequence: Sequence,
    block: AbstractBlock,
    config: Configuration,
    qc_register: Register,
) -> None:
    # we need this because of the case with a single type of block in a KronBlock
    # TODO: document properly

    n_qubits = len(sequence.register.qubits)

    # define qubit support
    qubit_support = block.qubit_support
    if not isinstance(qubit_support[0], int):
        qubit_support = tuple(range(n_qubits))

    if isinstance(block, AnalogBlock) and qc_register.device_specs.interaction != Interaction.NN:
        raise ValueError(f"Pulser does not support other interactions than '{Interaction.NN}'")

    local_channel = sequence.device.channels["rydberg_local"]
    global_channel = sequence.device.channels["rydberg_global"]

    rx = partial(digital_xy_rot_pulse, channel=local_channel, phase=0, config=config)
    ry = partial(digital_xy_rot_pulse, channel=local_channel, phase=PI / 2, config=config)
    rz = partial(digital_z_rot_pulse, channel=local_channel, phase=PI / 2, config=config)

    # TODO: lets move those to `@singledipatch`ed functions
    if isinstance(block, InteractionBlock):
        if not block.add_pattern:
            logger.warning(
                "Found block with `add_pattern = False`. This is not yet supported in the Pulser "
                "backend. If an addressing pattern is specified, it will be added to all blocks."
            )
        # Apply empty pulse if it's a global InteractionBlock
        if block.qubit_support.is_global:
            (uuid, duration) = block.parameters.uuid_param("duration")
            t = evaluate(duration) if duration.is_number else sequence.declare_variable(uuid)
            pulse = Pulse.ConstantPulse(duration=t, amplitude=0, detuning=0, phase=0)
            sequence.add(pulse, GLOBAL_CHANNEL, "wait-for-all")

        # Do nothing if its a non-global InteractionBlock, because that means
        # we are doing a rotation on other qubits which will include the interaction term
        else:
            support = set(block.qubit_support)
            if not support.issubset(sequence.register.qubits):
                raise ValueError("Trying evolve the interaction on qubits outside of support.")

    elif isinstance(block, ConstantAnalogRotation):
        if not block.add_pattern:
            logger.warning(
                "Found block with `add_pattern = False`. This is not yet supported in the Pulser "
                "backend. If an addressing pattern is specified, it will be added to all blocks."
            )
        ps = block.parameters
        (a_uuid, alpha) = ps.uuid_param("alpha")
        (w_uuid, omega) = ps.uuid_param("omega")
        (p_uuid, phase) = ps.uuid_param("phase")
        (d_uuid, detuning) = ps.uuid_param("delta")

        a = evaluate(alpha) if alpha.is_number else sequence.declare_variable(a_uuid)
        w = evaluate(omega) if omega.is_number else sequence.declare_variable(w_uuid)
        p = evaluate(phase) if phase.is_number else sequence.declare_variable(p_uuid)
        d = evaluate(detuning) if detuning.is_number else sequence.declare_variable(d_uuid)

        # calculate generator eigenvalues
        block.eigenvalues_generator = block.compute_eigenvalues_generator(block, qc_register)

        if block.qubit_support.is_global:
            pulse = analog_rot_pulse(a, w, p, d, global_channel, config)
            sequence.add(pulse, GLOBAL_CHANNEL, protocol="wait-for-all")
        else:
            pulse = analog_rot_pulse(a, w, p, d, local_channel, config)
            sequence.target(qubit_support, LOCAL_CHANNEL)
            sequence.add(pulse, LOCAL_CHANNEL, protocol="wait-for-all")

    elif isinstance(block, AnalogEntanglement):
        (uuid, duration) = block.parameters.uuid_param("duration")
        t = evaluate(duration) if duration.is_number else sequence.declare_variable(uuid)
        sequence.add(
            entangle_pulse(t, global_channel, config), GLOBAL_CHANNEL, protocol="wait-for-all"
        )

    elif isinstance(block, (RX, RY, RZ)):
        (uuid, p) = block.parameters.uuid_param("parameter")
        angle = evaluate(p) if p.is_number else sequence.declare_variable(uuid)
        if isinstance(block, RX):
            pulse = rx(angle)
        elif isinstance(block, RY):
            pulse = ry(angle)
        elif isinstance(block, RZ):
            pulse = rz(angle)
        sequence.target(qubit_support, LOCAL_CHANNEL)
        sequence.add(pulse, LOCAL_CHANNEL, protocol="wait-for-all")

    elif isinstance(block, CompositeBlock) or isinstance(block, AnalogComposite):
        for block in block.blocks:
            add_pulses(sequence, block, config, qc_register)

    else:
        msg = f"The pulser backend currently does not support blocks of type: {type(block)}"
        raise NotImplementedError(msg)


def analog_rot_pulse(
    alpha: TVar | float,
    omega: TVar | float,
    phase: TVar | float,
    detuning: TVar | float,
    channel: Channel,
    config: Configuration | None = None,
) -> Pulse:
    # omega in rad/us; detuning in rad/us
    if config is not None:
        if channel.addressing == "Global":
            max_amp = config.amplitude_global if config.amplitude_global is not None else omega
        elif channel.addressing == "Local":
            max_amp = config.amplitude_local if config.amplitude_local is not None else omega
        max_det = config.detuning if config.detuning is not None else detuning
    else:
        max_amp = omega
        max_det = detuning

    # get pulse duration in ns
    duration = 1000 * abs(alpha) / np.sqrt(omega**2 + detuning**2)

    # create amplitude waveform
    amp_wf = SquareWaveform.from_duration(
        duration=duration,  # type: ignore
        max_amp=max_amp,  # type: ignore[arg-type]
        duration_steps=channel.clock_period,  # type: ignore[attr-defined]
        min_duration=channel.min_duration,
    )

    # create detuning waveform
    det_wf = SquareWaveform.from_duration(
        duration=duration,  # type: ignore
        max_amp=max_det,  # type: ignore[arg-type]
        duration_steps=channel.clock_period,  # type: ignore[attr-defined]
        min_duration=channel.min_duration,
    )

    return Pulse(amplitude=amp_wf, detuning=det_wf, phase=abs(phase))


def entangle_pulse(
    duration: TVar | float, channel: Channel, config: Configuration | None = None
) -> Pulse:
    if config is None:
        max_amp = channel.max_amp
    else:
        max_amp = (
            config.amplitude_global if config.amplitude_global is not None else channel.max_amp
        )

    clock = channel.clock_period
    delay_wf = ConstantWaveform(clock * np.ceil(duration / clock), 0)  # type: ignore
    half_pi_wf = SquareWaveform.from_area(
        area=PI / 2,
        max_amp=max_amp,  # type: ignore[arg-type]
        duration_steps=clock,  # type: ignore[attr-defined]
        min_duration=channel.min_duration,
    )

    detuning_wf = RampWaveform(duration=half_pi_wf.duration, start=0, stop=PI)
    amplitude = CompositeWaveform(half_pi_wf, delay_wf)
    detuning = CompositeWaveform(detuning_wf, delay_wf)
    return Pulse(amplitude=amplitude, detuning=detuning, phase=PI / 2)


def digital_xy_rot_pulse(
    angle: TVar | float, phase: float, channel: Channel, config: Configuration | None = None
) -> Pulse:
    if config is None:
        max_amp = channel.max_amp
    else:
        max_amp = config.amplitude_local if config.amplitude_local is not None else channel.max_amp

    # TODO: Implement reverse rotation for angles bigger than π
    amplitude_wf = SquareWaveform.from_area(
        area=abs(angle),  # type: ignore
        max_amp=max_amp,  # type: ignore[arg-type]
        duration_steps=channel.clock_period,  # type: ignore[attr-defined]
        min_duration=channel.min_duration,
    )

    return Pulse.ConstantDetuning(amplitude=amplitude_wf, detuning=0, phase=phase)


def digital_z_rot_pulse(
    angle: TVar | float, phase: float, channel: Channel, config: Configuration | None = None
) -> Pulse:
    if config is None:
        max_det = channel.max_abs_detuning
    else:
        max_det = config.detuning if config.detuning is not None else channel.max_abs_detuning

    # get pulse duration in ns
    duration = 1000 * abs(angle) / max_det

    # create amplitude waveform
    amp_wf = SquareWaveform.from_duration(
        duration=duration,  # type: ignore
        max_amp=0.0,  # type: ignore[arg-type]
        duration_steps=channel.clock_period,  # type: ignore[attr-defined]
        min_duration=channel.min_duration,
    )

    # create detuning waveform
    det_wf = SquareWaveform.from_duration(
        duration=duration,  # type: ignore
        max_amp=max_det,  # type: ignore[arg-type]
        duration_steps=channel.clock_period,  # type: ignore[attr-defined]
        min_duration=channel.min_duration,
    )

    return Pulse(amplitude=amp_wf, detuning=det_wf, phase=abs(phase))



from __future__ import annotations

import importlib
from collections import Counter
from dataclasses import dataclass
from typing import Callable, cast

from qadence.noise.protocols import Noise

PROTOCOL_TO_MODULE = {
    "readout": "qadence.mitigations.readout",
    "zne": "qadence.mitigations.analog_zne",
}


@dataclass
class Mitigations:
    READOUT = "readout"
    ANALOG_ZNE = "zne"

    def __init__(self, protocol: str, options: dict = dict()) -> None:
        self.protocol: str = protocol
        self.options: dict = options

    def get_mitigation_fn(self) -> Callable:
        try:
            module = importlib.import_module(PROTOCOL_TO_MODULE[self.protocol])
        except KeyError:
            ImportError(f"The module corresponding to the protocol {self.protocol} is not found.")
        fn = getattr(module, "mitigate")
        return cast(Callable, fn)

    def _to_dict(self) -> dict:
        return {"protocol": self.protocol, "options": self.options}

    @classmethod
    def _from_dict(cls, d: dict) -> Mitigations | None:
        if d:
            return cls(d["protocol"], **d["options"])
        return None

    @classmethod
    def list(cls) -> list:
        return list(filter(lambda el: not el.startswith("__"), dir(cls)))


def apply_mitigation(
    noise: Noise, mitigation: Mitigations, samples: list[Counter]
) -> list[Counter]:
    """Apply mitigation to samples."""
    mitigation_fn = mitigation.get_mitigation_fn()
    mitigated_samples: list[Counter] = mitigation_fn(
        noise=noise, mitigation=mitigation, samples=samples
    )
    return mitigated_samples



from __future__ import annotations

from qadence.blocks.utils import chain
from qadence.circuit import QuantumCircuit
from qadence.operations import I


def fill_identities(circ: QuantumCircuit) -> QuantumCircuit:
    empty_wires = set(range(circ.n_qubits)) - set(circ.block.qubit_support)
    if len(empty_wires) > 0:
        ids = chain(I(i) for i in empty_wires)
        return QuantumCircuit(circ.n_qubits, chain(circ.block, ids))
    return circ



from __future__ import annotations

from copy import deepcopy
from itertools import product
from math import dist
from typing import Any

import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
from deepdiff import DeepDiff
from networkx.classes.reportviews import EdgeView, NodeView

from qadence.analog import IdealDevice, RydbergDevice
from qadence.types import LatticeTopology

# Modules to be automatically added to the qadence namespace
__all__ = ["Register"]


DEFAULT_DEVICE = IdealDevice()


def _scale_node_positions(graph: nx.Graph, min_distance: float, spacing: float) -> None:
    scaled_nodes = {}
    scale_factor = spacing / min_distance
    for k, node in graph.nodes.items():
        (x, y) = node["pos"]
        scaled_nodes[k] = {"pos": (x * scale_factor, y * scale_factor)}
    nx.set_node_attributes(graph, scaled_nodes)


class Register:
    def __init__(
        self,
        support: nx.Graph | int,
        spacing: float | None = 1.0,
        device_specs: RydbergDevice = DEFAULT_DEVICE,
    ):
        """
        A register of qubits including 2D coordinates.

        Instantiating the Register class directly is only recommended for building custom registers.
        For most uses where a predefined lattice is desired it is recommended to use the various
        class methods available, e.g. `Register.triangular_lattice`.

        Arguments:
            support: A NetworkX graph or number of qubits. Nodes can include a `"pos"` attribute
                such that e.g.: `graph.nodes = {0: {"pos": (2,3)}, 1: {"pos": (0,0)}, ...}` which
                will be used in backends that need qubit coordinates. Passing a number of qubits
                calls `Register.all_to_all(n_qubits)`.
            spacing: Value set as the distance between the two closest qubits. The spacing
                argument is also available for all the class method constructors.

        Examples:
        ```python exec="on" source="material-block"
        from qadence import Register

        reg_all = Register.all_to_all(n_qubits = 4)
        reg_line = Register.line(n_qubits = 4)
        reg_circle = Register.circle(n_qubits = 4)
        reg_squre = Register.square(qubits_side = 2)
        reg_rect = Register.rectangular_lattice(qubits_row = 2, qubits_col = 2)
        reg_triang = Register.triangular_lattice(n_cells_row = 2, n_cells_col = 2)
        reg_honey = Register.honeycomb_lattice(n_cells_row = 2, n_cells_col = 2)
        ```
        """
        if device_specs is not None and not isinstance(device_specs, RydbergDevice):
            raise ValueError("Device specs are not valid. Please pass a `RydbergDevice` instance.")

        self.device_specs = device_specs

        self.graph = support if isinstance(support, nx.Graph) else alltoall_graph(support)

        if spacing is not None and self.min_distance != 0.0:
            _scale_node_positions(self.graph, self.min_distance, spacing)

    @property
    def n_qubits(self) -> int:
        """Total number of qubits in the register."""
        return len(self.graph)

    @classmethod
    def from_coordinates(
        cls,
        coords: list[tuple],
        lattice: LatticeTopology | str = LatticeTopology.ARBITRARY,
        spacing: float | None = None,
        device_specs: RydbergDevice = DEFAULT_DEVICE,
    ) -> Register:
        """
        Build a register from a list of qubit coordinates.

        Each node is added to the underlying
        graph with the respective coordinates, but the edges are left empty.

        Arguments:
            coords: List of qubit coordinate tuples.
        """
        graph = nx.Graph()
        for i, pos in enumerate(coords):
            graph.add_node(i, pos=pos)
        return cls(graph, spacing, device_specs)

    @classmethod
    def line(
        cls,
        n_qubits: int,
        spacing: float = 1.0,
        device_specs: RydbergDevice = DEFAULT_DEVICE,
    ) -> Register:
        """
        Build a line register.

        Arguments:
            n_qubits: Total number of qubits.
        """
        return cls(line_graph(n_qubits), spacing, device_specs)

    @classmethod
    def circle(
        cls,
        n_qubits: int,
        spacing: float = 1.0,
        device_specs: RydbergDevice = DEFAULT_DEVICE,
    ) -> Register:
        """
        Build a circle register.

        Arguments:
            n_qubits: Total number of qubits.
        """
        graph = nx.grid_2d_graph(n_qubits, 1, periodic=True)
        graph = nx.relabel_nodes(graph, {(i, 0): i for i in range(n_qubits)})
        coords = nx.circular_layout(graph)
        values = {i: {"pos": pos} for i, pos in coords.items()}
        nx.set_node_attributes(graph, values)
        return cls(graph, spacing, device_specs)

    @classmethod
    def square(
        cls,
        qubits_side: int,
        spacing: float = 1.0,
        device_specs: RydbergDevice = DEFAULT_DEVICE,
    ) -> Register:
        """
        Build a square register.

        Arguments:
            qubits_side: Number of qubits on one side of the square.
        """
        n_points = 4 * (qubits_side - 1)

        def gen_points() -> np.ndarray:
            rotate_left = np.array([[0.0, -1.0], [1.0, 0.0]])
            increment = np.array([0.0, 1.0])

            points = [np.array([0.0, 0.0])]
            counter = 1
            while len(points) < n_points:
                points.append(points[-1] + increment)

                counter = (counter + 1) % qubits_side
                if counter == 0:
                    increment = rotate_left.dot(increment)
                    counter = 1
            points = np.array(points)  # type: ignore[assignment]
            points -= np.mean(points, axis=0)

            return points  # type: ignore[return-value]

        graph = nx.grid_2d_graph(n_points, 1, periodic=True)
        graph = nx.relabel_nodes(graph, {(i, 0): i for i in range(n_points)})
        values = {i: {"pos": point} for i, point in zip(graph.nodes, gen_points())}
        nx.set_node_attributes(graph, values)
        return cls(graph, spacing, device_specs)

    @classmethod
    def all_to_all(
        cls,
        n_qubits: int,
        spacing: float = 1.0,
        device_specs: RydbergDevice = DEFAULT_DEVICE,
    ) -> Register:
        """
        Build a register with an all-to-all connectivity graph.

        The graph is projected
        onto a 2D space and the qubit coordinates are set using a spring layout algorithm.

        Arguments:
            n_qubits: Total number of qubits.
        """
        return cls(alltoall_graph(n_qubits), spacing, device_specs)

    @classmethod
    def rectangular_lattice(
        cls,
        qubits_row: int,
        qubits_col: int,
        spacing: float = 1.0,
        device_specs: RydbergDevice = DEFAULT_DEVICE,
    ) -> Register:
        graph = nx.grid_2d_graph(qubits_col, qubits_row)
        values = {i: {"pos": node} for (i, node) in enumerate(graph.nodes)}
        graph = nx.relabel_nodes(graph, {(i, j): k for k, (i, j) in enumerate(graph.nodes)})
        nx.set_node_attributes(graph, values)
        """
        Build a rectangular lattice register.

        Arguments:
            qubits_row: Number of qubits in each row.
            qubits_col: Number of qubits in each column.
        """
        return cls(graph, spacing, device_specs)

    @classmethod
    def triangular_lattice(
        cls,
        n_cells_row: int,
        n_cells_col: int,
        spacing: float = 1.0,
        device_specs: RydbergDevice = DEFAULT_DEVICE,
    ) -> Register:
        """
        Build a triangular lattice register.

        Each cell is a triangle made up of three qubits.

        Arguments:
            n_cells_row: Number of cells in each row.
            n_cells_col: Number of cells in each column.
        """
        return cls(triangular_lattice_graph(n_cells_row, n_cells_col), spacing, device_specs)

    @classmethod
    def honeycomb_lattice(
        cls,
        n_cells_row: int,
        n_cells_col: int,
        spacing: float = 1.0,
        device_specs: RydbergDevice = DEFAULT_DEVICE,
    ) -> Register:
        """
        Build a honeycomb lattice register.

        Each cell is an hexagon made up of six qubits.

        Arguments:
            n_cells_row: Number of cells in each row.
            n_cells_col: Number of cells in each column.
        """
        graph = nx.hexagonal_lattice_graph(n_cells_row, n_cells_col)
        graph = nx.relabel_nodes(graph, {(i, j): k for k, (i, j) in enumerate(graph.nodes)})
        return cls(graph, spacing, device_specs)

    @classmethod
    def lattice(cls, topology: LatticeTopology | str, *args: Any, **kwargs: Any) -> Register:
        return getattr(cls, topology)(*args, **kwargs)  # type: ignore[no-any-return]

    def draw(self, show: bool = True) -> None:
        """Draw the underlying NetworkX graph representing the register."""
        coords = {i: n["pos"] for i, n in self.graph.nodes.items()}
        nx.draw(self.graph, with_labels=True, pos=coords)
        if show:
            plt.gcf().show()

    def __getitem__(self, item: int) -> Any:
        return self.graph.nodes[item]

    @property
    def nodes(self) -> NodeView:
        """Return the NodeView of the underlying NetworkX graph."""
        return self.graph.nodes

    @property
    def edges(self) -> EdgeView:
        """Return the EdgeView of the underlying NetworkX graph."""
        return self.graph.edges

    @property
    def support(self) -> set:
        """Return the set of qubits in the register."""
        return set(self.nodes)

    @property
    def coords(self) -> dict:
        """Return the dictionary of qubit coordinates."""
        return {i: tuple(node.get("pos", ())) for i, node in self.nodes.items()}

    @property
    def all_node_pairs(self) -> EdgeView:
        """Return a list of all possible qubit pairs in the register."""
        return list(filter(lambda x: x[0] < x[1], product(self.support, self.support)))

    @property
    def distances(self) -> dict:
        """Return a dictionary of distances for all qubit pairs in the register."""
        coords = self.coords
        return {edge: dist(coords[edge[0]], coords[edge[1]]) for edge in self.all_node_pairs}

    @property
    def edge_distances(self) -> dict:
        """
        Return a dictionary of distances for the qubit pairs that are.

        connected by an edge in the underlying NetworkX graph.
        """
        coords = self.coords
        return {edge: dist(coords[edge[0]], coords[edge[1]]) for edge in self.edges}

    @property
    def min_distance(self) -> float:
        """Return the minimum distance between two qubts in the register."""
        distances = self.distances
        value: float = min(self.distances.values()) if len(distances) > 0 else 0.0
        return value

    def rescale_coords(self, scaling: float) -> Register:
        """
        Rescale the coordinates of all qubits in the register.

        Arguments:
            scaling: Scaling value.
        """
        g = deepcopy(self.graph)
        _scale_node_positions(g, min_distance=1.0, spacing=scaling)
        return Register(g, spacing=None, device_specs=self.device_specs)

    def _to_dict(self) -> dict:
        return {
            "graph": nx.node_link_data(self.graph),
            "device_specs": self.device_specs._to_dict(),
        }

    @classmethod
    def _from_dict(cls, d: dict) -> Register:
        device_dict = d.get("device_specs", None)
        if device_dict is None:
            device_dict = DEFAULT_DEVICE._to_dict()

        return cls(
            support=nx.node_link_graph(d["graph"]),
            device_specs=RydbergDevice._from_dict(device_dict),
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Register):
            return False
        return (
            DeepDiff(self.coords, other.coords, ignore_order=False) == {}
            and nx.is_isomorphic(self.graph, other.graph)
            and self.n_qubits == other.n_qubits
        )


def line_graph(n_qubits: int) -> nx.Graph:
    graph = nx.Graph()
    for i in range(n_qubits):
        graph.add_node(i, pos=(i, 0.0))
    for i, j in zip(range(n_qubits - 1), range(1, n_qubits)):
        graph.add_edge(i, j)
    return graph


def triangular_lattice_graph(n_cells_row: int, n_cells_col: int) -> nx.Graph:
    graph = nx.triangular_lattice_graph(n_cells_row, n_cells_col)
    graph = nx.relabel_nodes(graph, {(i, j): k for k, (i, j) in enumerate(graph.nodes)})
    return graph


def alltoall_graph(n_qubits: int) -> nx.Graph:
    if n_qubits == 2:
        return line_graph(2)
    if n_qubits == 3:
        return triangular_lattice_graph(1, 1)

    graph = nx.complete_graph(n_qubits)
    # set seed to make sure the produced graphs are reproducible
    coords = nx.spring_layout(graph, seed=0)
    for i, pos in coords.items():
        graph.nodes[i]["pos"] = tuple(pos)
    return graph



from __future__ import annotations

from typing import Any, Callable

from qadence.blocks import AbstractBlock, CompositeBlock, add, chain, kron
from qadence.blocks.analog import AnalogChain

COMPOSE_FN_DICT = {
    "ChainBlock": chain,
    "AnalogChain": chain,
    "KronBlock": kron,
    "AnalogKron": kron,
    "AddBlock": add,
}


def apply_fn_to_blocks(
    input_block: AbstractBlock, block_fn: Callable, *args: Any, **kwargs: Any
) -> AbstractBlock:
    """
    Recurses through the block tree and applies a given function to all the leaf blocks.

    Arguments:
        input_block: tree of blocks on which to apply the recurse.
        block_fn: callable function to apply to each leaf block.
        args: any positional arguments to pass to the leaf function.
        kwargs: any keyword arguments to pass to the leaf function.
    """

    if isinstance(input_block, (CompositeBlock, AnalogChain)):
        parsed_blocks = [
            apply_fn_to_blocks(block, block_fn, *args, **kwargs) for block in input_block.blocks
        ]
        compose_fn = COMPOSE_FN_DICT[type(input_block).__name__]
        output_block = compose_fn(*parsed_blocks)  # type: ignore [arg-type]
    else:
        # AnalogKrons are considered as a leaf block
        output_block = block_fn(input_block, *args, **kwargs)

    return output_block



from __future__ import annotations

import math
from collections import Counter
from functools import partial
from logging import getLogger
from typing import TYPE_CHECKING, Any

import numpy as np
import sympy
from numpy.typing import ArrayLike
from torch import Tensor, stack, vmap
from torch import complex as make_complex
from torch.linalg import eigvals

from qadence.types import Endianness, ResultType, TNumber

if TYPE_CHECKING:
    from qadence.operations import Projector


# Modules to be automatically added to the qadence namespace
__all__ = []  # type: ignore


logger = getLogger(__name__)


def basis_to_int(basis: str, endianness: Endianness = Endianness.BIG) -> int:
    """
    Converts a computational basis state to an int.

    - `endianness = "Big"` reads the most significant bit in qubit 0 (leftmost).
    - `endianness = "Little"` reads the least significant bit in qubit 0 (leftmost).

    Arguments:
        basis (str): A computational basis state.
        endianness (Endianness): The Endianness when reading the basis state.

    Returns:
        The corresponding integer.

    Examples:
    ```python exec="on" source="material-block" result="json"
    from qadence.utils import basis_to_int, Endianness

    k = basis_to_int(basis="10", endianness=Endianness.BIG)
    print(k)
    ```
    """
    if endianness == Endianness.BIG:
        return int(basis, 2)
    else:
        return int(basis[::-1], 2)


def int_to_basis(
    k: int, n_qubits: int | None = None, endianness: Endianness = Endianness.BIG
) -> str:
    """
    Converts an integer to its corresponding basis state.

    - `endianness = "Big"` stores the most significant bit in qubit 0 (leftmost).
    - `endianness = "Little"` stores the least significant bit in qubit 0 (leftmost).

    Arguments:
        k (int): The int to convert.
        n_qubits (int): The total number of qubits in the basis state.
        endianness (Endianness): The Endianness of the resulting basis state.

    Returns:
        A computational basis state.

    Examples:
    ```python exec="on" source="material-block" result="json"
    from qadence.utils import int_to_basis, Endianness

    bs = int_to_basis(k=1, n_qubits=2, endianness=Endianness.BIG)
    print(bs)
    ```
    """
    if n_qubits is None:
        n_qubits = int(math.log(k + 0.6) / math.log(2)) + 1
    assert k <= 2**n_qubits - 1, "k can not be larger than 2**n_qubits-1."
    basis = format(k, "0{}b".format(n_qubits))
    if endianness == Endianness.BIG:
        return basis
    else:
        return basis[::-1]


def nqubits_to_basis(
    n_qubits: int,
    result_type: ResultType = ResultType.STRING,
    endianness: Endianness = Endianness.BIG,
) -> list[str] | Tensor | np.array:
    """
    Creates all basis states for a given number of qubits, endianness and format.

    Arguments:
        n_qubits: The total number of qubits.
        result_type: The data type of the resulting states.
        endianness: The Endianness of the resulting states.

    Returns:
        The full computational basis for n_qubits.

    Examples:
    ```python exec="on" source="material-block" result="json"
    from qadence.utils import nqubits_to_basis, Endianness, ResultType
    basis_type = ResultType.Torch
    bs = nqubits_to_basis(n_qubits=2, result_type= basis_type, endianness=Endianness.BIG)
    print(bs)
    ```
    """
    basis_strings = [int_to_basis(k, n_qubits, endianness) for k in range(0, 2**n_qubits)]
    if result_type == ResultType.STRING:
        return basis_strings
    else:
        basis_list = [list(map(int, tuple(basis))) for basis in basis_strings]
        if result_type == ResultType.TORCH:
            return stack([Tensor(basis) for basis in basis_list])
        elif result_type == ResultType.NUMPY:
            return np.stack([np.array(basis) for basis in basis_list])


def samples_to_integers(samples: Counter, endianness: Endianness = Endianness.BIG) -> Counter:
    """
    Converts a Counter of basis state samples to integer values.

    Args:
        samples (Counter({bits: counts})): basis state sample counter.
        endianness (Endianness): endianness to use for conversion.

    Returns:
        Counter({ints: counts}): samples converted
    """

    return Counter({basis_to_int(k, endianness): v for k, v in samples.items()})


def format_number(x: float | complex, num_digits: int = 3) -> str:
    if isinstance(x, int):
        return f"{x}"
    elif isinstance(x, float):
        return f"{x:.{num_digits}f}"
    elif isinstance(x, complex):
        re = "" if np.isclose(x.real, 0) else f"{x.real:.{num_digits}f}"
        im = "" if np.isclose(x.imag, 0) else f"{x.imag:.{num_digits}f}"
        if len(re) > 0 and len(im) > 0:
            return f"{re}+{im}j"
        elif len(re) > 0 and len(im) == 0:
            return re
        elif len(re) == 0 and len(im) > 0:
            return f"{im}j"
        else:
            return "0"
    else:
        raise ValueError(f"Unknown number type: {type(x)}")


def format_parameter(p: sympy.Basic, num_digits: int = 3) -> str:
    """Format numerical values within a sympy expression."""

    def round_expr(expr: sympy.Basic, num_digits: int) -> sympy.Basic:
        return expr.xreplace({n: round(n, num_digits) for n in expr.atoms(sympy.Number)})

    expr = round_expr(p, num_digits)
    conv_str = str(expr)
    if expr.is_real and len(conv_str) > num_digits + 2:
        conv_str = conv_str[0 : num_digits + 2]
    return conv_str


def print_sympy_expr(expr: sympy.Expr, num_digits: int = 3) -> str:
    """
    Converts numerical values in a sympy expression.

    The result is a numerical expression with fewer digits for better readability.
    """
    from qadence.parameters import sympy_to_numeric

    round_dict = {sympy_to_numeric(n): round(n, num_digits) for n in expr.atoms(sympy.Number)}
    return str(expr.xreplace(round_dict))


def isclose(
    x: TNumber | Any, y: TNumber | Any, rel_tol: float = 1e-5, abs_tol: float = 1e-07
) -> bool:
    if isinstance(x, complex) or isinstance(y, complex):
        return abs(x - y) <= max(rel_tol * max(abs(x), abs(y)), abs_tol)  # type: ignore

    return math.isclose(x, y, rel_tol=rel_tol, abs_tol=abs_tol)  # type: ignore


def eigenvalues(
    x: Tensor, max_num_evals: int | None = None, max_num_gaps: int | None = None
) -> Tensor:
    # FIXME: Currently doing full decomposition everytime because it is generally faster
    # than converting to numpy and using the scipy routines to select first k-eigenvalues.
    # Furthermore, we should exploit the matrices being Hermitian.

    # get all eigenvalues of generator
    eigenspectrum = eigvals(x).squeeze(0)

    if max_num_evals and not max_num_gaps:
        # get specified number of eigenvalues of generator
        eigenvals = eigenspectrum[:max_num_evals]
    elif max_num_gaps and not max_num_evals:
        # get eigenvalues of generator corresponding to specified number of spectral gaps
        k = int(np.ceil(0.5 * (1 + np.sqrt(1 + 8 * max_num_gaps))))
        eigenvals = eigenspectrum[:k]
    else:
        eigenvals = eigenspectrum

    return eigenvals


def _round_complex(t: Tensor, decimals: int = 4) -> Tensor:
    def _round(_t: Tensor) -> Tensor:
        r = _t.real.round(decimals=decimals)
        i = _t.imag.round(decimals=decimals)
        return make_complex(r, i)

    fn = vmap(_round)
    return fn(t)


def is_qadence_shape(state: ArrayLike, n_qubits: int) -> bool:
    if len(state.shape) < 2:
        raise ValueError(
            f"Provided state is required to have atleast two dimensions. Got shape {state.shape}"
        )
    return state.shape[1] == 2**n_qubits  # type: ignore[no-any-return]


def validate_values_and_state(
    state: ArrayLike | None, n_qubits: int, param_values: dict[str, Tensor] = None
) -> None:
    if state is not None:
        from qadence.backends.utils import infer_batchsize

        if isinstance(state, Tensor):
            if state is not None:
                batch_size_state = (
                    state.shape[0] if is_qadence_shape(state, n_qubits=n_qubits) else state.size(-1)
                )
                batch_size_values = infer_batchsize(param_values)
                if batch_size_state != batch_size_values and (
                    batch_size_values > 1 and batch_size_state > 1
                ):
                    raise ValueError(
                        "Batching of parameter values and states is only valid for the cases:\
                                    (1) batch_size_values == batch_size_state\
                                    (2) batch_size_values == 1 and batch_size_state > 1\
                                    (3) batch_size_values > 1 and batch_size_state == 1."
                    )
        else:
            if not is_qadence_shape(state, n_qubits) or state.shape[0] > 1:
                raise ValueError("Jax only supports unbatched states.")


def one_qubit_projector(state: str, target: int) -> Projector:
    """Returns the projector for a single qubit system.

    Args:
        state (str): The state of the projector.
        target (int): The target qubit.

    Returns:
        Projector: The projector operator.
    """
    from qadence.operations import Projector

    assert state in ["0", "1"], "State must be either '0' or '1'."
    return Projector(ket=state, bra=state, qubit_support=target)


def one_qubit_projector_matrix(state: str) -> Tensor:
    """Returns the projector for a single qubit system.

    Args:
        state (str): The state of the projector.

    Returns:
        Tensor: The projector operator.
    """
    return one_qubit_projector(state, 0).tensor().squeeze()


P0 = partial(one_qubit_projector, "0")
P1 = partial(one_qubit_projector, "1")
P0_MATRIX = one_qubit_projector_matrix("0")
P1_MATRIX = one_qubit_projector_matrix("1")



from __future__ import annotations

from collections import Counter
from functools import reduce

import numpy as np
import torch
from torch import Tensor

from qadence.backend import Backend
from qadence.backends.pyqtorch import Backend as PyQBackend
from qadence.blocks import AbstractBlock, chain, kron
from qadence.blocks.block_to_tensor import HMAT, IMAT, SDAGMAT, ZMAT, block_to_tensor
from qadence.blocks.composite import CompositeBlock
from qadence.blocks.primitive import PrimitiveBlock
from qadence.blocks.utils import get_pauli_blocks, unroll_block_with_scaling
from qadence.circuit import QuantumCircuit
from qadence.engines.differentiable_backend import DifferentiableBackend
from qadence.noise import Noise
from qadence.operations import X, Y, Z
from qadence.types import Endianness
from qadence.utils import P0_MATRIX, P1_MATRIX

pauli_gates = [X, Y, Z]


UNITARY_TENSOR = [
    ZMAT @ HMAT,
    SDAGMAT.squeeze(dim=0) @ HMAT,
    IMAT,
]


def identity(n_qubits: int) -> Tensor:
    return torch.eye(2**n_qubits, dtype=torch.complex128)


def _max_observable_weight(observable: AbstractBlock) -> int:
    """
    Get the maximal weight for the given observable.

    The weight is a measure of the locality of the observable,
    a count of the number of qubits on which the observable acts
    non-trivially.

    See https://arxiv.org/pdf/2002.08953.pdf
    Supplementary Material 1 and Eq. (S17).
    """
    pauli_decomposition = unroll_block_with_scaling(observable)
    weights = []
    for pauli_term in pauli_decomposition:
        weight = 0
        block = pauli_term[0]
        if isinstance(block, PrimitiveBlock):
            if isinstance(block, (X, Y, Z)):
                weight += 1
            weights.append(weight)
        else:
            pauli_blocks = get_pauli_blocks(block=block)
            weight = 0
            for block in pauli_blocks:
                if isinstance(block, (X, Y, Z)):
                    weight += 1
            weights.append(weight)
    return max(weights)


def maximal_weight(observables: list[AbstractBlock]) -> int:
    """Return the maximal weight if a list of observables is provided."""
    return max([_max_observable_weight(observable=observable) for observable in observables])


def number_of_samples(
    observables: list[AbstractBlock], accuracy: float, confidence: float
) -> tuple[int, ...]:
    """
    Estimate an optimal shot budget and a shadow partition size.

    to guarantee given accuracy on all observables expectation values
    within 1 - confidence range.

    See https://arxiv.org/pdf/2002.08953.pdf
    Supplementary Material 1 and Eqs. (S23)-(S24).
    """
    max_k = maximal_weight(observables=observables)
    N = round(3**max_k * 34.0 / accuracy**2)
    K = round(2.0 * np.log(2.0 * len(observables) / confidence))
    return N, K


def local_shadow(sample: Counter, unitary_ids: list) -> Tensor:
    """
    Compute local shadow by inverting the quantum channel for each projector state.

    See https://arxiv.org/pdf/2002.08953.pdf
    Supplementary Material 1 and Eqs. (S17,S44).

    Expects a sample bitstring in ILO.
    """
    bitstring = list(sample.keys())[0]
    local_density_matrices = []
    for bit, unitary_id in zip(bitstring, unitary_ids):
        proj_mat = P0_MATRIX if bit == "0" else P1_MATRIX
        unitary_tensor = UNITARY_TENSOR[unitary_id].squeeze(dim=0)
        local_density_matrices.append(
            3 * (unitary_tensor.adjoint() @ proj_mat @ unitary_tensor) - identity(1)
        )
    if len(local_density_matrices) == 1:
        return local_density_matrices[0]
    else:
        return reduce(torch.kron, local_density_matrices)


def classical_shadow(
    shadow_size: int,
    circuit: QuantumCircuit,
    param_values: dict,
    state: Tensor | None = None,
    backend: Backend | DifferentiableBackend = PyQBackend(),
    # FIXME: Changed below from Little to Big, double-check when Roland is back
    noise: Noise | None = None,
    endianness: Endianness = Endianness.BIG,
) -> list:
    shadow: list = []
    # TODO: Parallelize embarrassingly parallel loop.
    for _ in range(shadow_size):
        unitary_ids = np.random.randint(0, 3, size=(1, circuit.n_qubits))[0]
        random_unitary = [
            pauli_gates[unitary_ids[qubit]](qubit) for qubit in range(circuit.n_qubits)
        ]
        if len(random_unitary) == 1:
            random_unitary_block = random_unitary[0]
        else:
            random_unitary_block = kron(*random_unitary)
        rotated_circuit = QuantumCircuit(
            circuit.n_qubits,
            chain(circuit.block, random_unitary_block),
        )
        # Reverse endianness to get sample bitstrings in ILO.
        conv_circ = backend.circuit(rotated_circuit)
        samples = backend.sample(
            circuit=conv_circ,
            param_values=param_values,
            n_shots=1,
            state=state,
            noise=noise,
            endianness=endianness,
        )
        batched_shadow = []
        for batch in samples:
            batched_shadow.append(local_shadow(sample=batch, unitary_ids=unitary_ids))
        shadow.append(batched_shadow)

    # Reshape the shadow by batches of samples instead of samples of batches.
    # FIXME: Improve performance.
    return [list(s) for s in zip(*shadow)]


def reconstruct_state(shadow: list) -> Tensor:
    """Reconstruct the state density matrix for the given shadow."""
    return reduce(torch.add, shadow) / len(shadow)


def compute_traces(
    qubit_support: tuple,
    N: int,
    K: int,
    shadow: list,
    observable: AbstractBlock,
    endianness: Endianness = Endianness.BIG,
) -> list:
    floor = int(np.floor(N / K))
    traces = []
    # TODO: Parallelize embarrassingly parallel loop.
    for k in range(K):
        reconstructed_state = reconstruct_state(shadow=shadow[k * floor : (k + 1) * floor])
        # Reshape the observable matrix to fit the density matrix dimensions
        # by filling indentites.
        # Please note the endianness is also flipped to get results in LE.
        # FIXME: Changed below from Little to Big, double-check when Roland is back
        # FIXME: Correct these comments.
        trace = (
            (
                block_to_tensor(
                    block=observable,
                    qubit_support=qubit_support,
                    endianness=Endianness.BIG,
                ).squeeze(dim=0)
                @ reconstructed_state
            )
            .trace()
            .real
        )
        traces.append(trace)
    return traces


def estimators(
    qubit_support: tuple,
    N: int,
    K: int,
    shadow: list,
    observable: AbstractBlock,
    endianness: Endianness = Endianness.BIG,
) -> Tensor:
    """
    Return estimators (traces of observable times mean density matrix).

    for K equally-sized shadow partitions.

    See https://arxiv.org/pdf/2002.08953.pdf
    Algorithm 1.
    """
    # If there is no Pauli-Z operator in the observable,
    # the sample can't "hit" that measurement.
    if isinstance(observable, PrimitiveBlock):
        if type(observable) == Z:
            traces = compute_traces(
                qubit_support=qubit_support,
                N=N,
                K=K,
                shadow=shadow,
                observable=observable,
                endianness=endianness,
            )
        else:
            traces = [torch.tensor(0.0)]
    elif isinstance(observable, CompositeBlock):
        if Z in observable:
            traces = compute_traces(
                qubit_support=qubit_support,
                N=N,
                K=K,
                shadow=shadow,
                observable=observable,
                endianness=endianness,
            )
        else:
            traces = [torch.tensor(0.0)]
    return torch.tensor(traces, dtype=torch.get_default_dtype())


def estimations(
    circuit: QuantumCircuit,
    observables: list[AbstractBlock],
    param_values: dict,
    shadow_size: int | None = None,
    accuracy: float = 0.1,
    confidence: float = 0.1,
    state: Tensor | None = None,
    backend: Backend | DifferentiableBackend = PyQBackend(),
    noise: Noise | None = None,
    endianness: Endianness = Endianness.BIG,
) -> Tensor:
    """Compute expectation values for all local observables using median of means."""
    # N is the estimated shot budget for the classical shadow to
    # achieve desired accuracy for all L = len(observables) within 1 - confidence probablity.
    # K is the size of the shadow partition.
    N, K = number_of_samples(observables=observables, accuracy=accuracy, confidence=confidence)
    if shadow_size is not None:
        N = shadow_size
    shadow = classical_shadow(
        shadow_size=N,
        circuit=circuit,
        param_values=param_values,
        state=state,
        backend=backend,
        noise=noise,
        endianness=endianness,
    )
    estimations = []
    for observable in observables:
        pauli_decomposition = unroll_block_with_scaling(observable)
        batch_estimations = []
        for batch in shadow:
            pauli_term_estimations = []
            for pauli_term in pauli_decomposition:
                # Get the estimators for the current Pauli term.
                # This is a tensor<float> of size K.
                estimation = estimators(
                    qubit_support=circuit.block.qubit_support,
                    N=N,
                    K=K,
                    shadow=batch,
                    observable=pauli_term[0],
                    endianness=endianness,
                )
                # Compute the median of means for the current Pauli term.
                # Weigh the median by the Pauli term scaling.
                pauli_term_estimations.append(torch.median(estimation) * pauli_term[1])
            # Sum the expectations for each Pauli term to get the expectation for the
            # current batch.
            batch_estimations.append(sum(pauli_term_estimations))
        estimations.append(batch_estimations)
    return torch.transpose(torch.tensor(estimations, dtype=torch.get_default_dtype()), 1, 0)


def compute_expectation(
    circuit: QuantumCircuit,
    observables: list[AbstractBlock],
    param_values: dict,
    options: dict,
    state: Tensor | None = None,
    backend: Backend | DifferentiableBackend = PyQBackend(),
    noise: Noise | None = None,
    endianness: Endianness = Endianness.BIG,
) -> Tensor:
    """
    Construct a classical shadow of a state to estimate observable expectation values.

    Args:
        circuit (QuantumCircuit): a circuit to prepare the state.
        observables (list[AbstractBlock]): a list of observables
            to estimate the expectation values from.
        param_values (dict): a dict of values to substitute the
            symbolic parameters for.
        options (dict): a dict of options for the measurement protocol.
            Here, shadow_size (int), accuracy (float) and confidence (float) are supported.
        state (Tensor | None): an initial input state.
        backend_name (BackendName): a backend name to retrieve computations from.
        noise: A noise model to use.
        endianness: Endianness of the observable estimate.

    Returns:
        expectations (Tensor): an estimation of the expectation values.
    """
    if not isinstance(observables, list):
        raise TypeError(
            "Observables must be of type <class 'List[AbstractBlock]'>. Got {}.".format(
                type(observables)
            )
        )
    shadow_size = options.get("shadow_size", None)
    accuracy = options.get("accuracy", None)
    if shadow_size is None and accuracy is None:
        KeyError(
            "Shadow protocol requires either an option"
            "'shadow_size' of type 'int' or 'accuracy' of type 'float'."
        )
    confidence = options.get("confidence", None)
    if confidence is None:
        KeyError("Shadow protocol requires a 'confidence' kwarg of type 'float'.")
    return estimations(
        circuit=circuit,
        observables=observables,
        param_values=param_values,
        shadow_size=shadow_size,
        accuracy=accuracy,
        confidence=confidence,
        state=state,
        backend=backend,
        noise=noise,
        endianness=endianness,
    )



from __future__ import annotations

from dataclasses import dataclass
from functools import singledispatch
from itertools import cycle
from typing import Any, Iterator

from torch import Tensor
from torch import device as torch_device
from torch.utils.data import DataLoader, IterableDataset, TensorDataset


@dataclass
class DictDataLoader:
    """This class only holds a dictionary of `DataLoader`s and samples from them."""

    dataloaders: dict[str, DataLoader]

    def __iter__(self) -> DictDataLoader:
        self.iters = {key: iter(dl) for key, dl in self.dataloaders.items()}
        return self

    def __next__(self) -> dict[str, Tensor]:
        return {key: next(it) for key, it in self.iters.items()}


class InfiniteTensorDataset(IterableDataset):
    def __init__(self, *tensors: Tensor):
        """Randomly sample points from the first dimension of the given tensors.

        Behaves like a normal torch `Dataset` just that we can sample from it as
        many times as we want.

        Examples:
        ```python exec="on" source="above" result="json"
        import torch
        from qadence.ml_tools.data import InfiniteTensorDataset

        x_data, y_data = torch.rand(5,2), torch.ones(5,1)
        # The dataset accepts any number of tensors with the same batch dimension
        ds = InfiniteTensorDataset(x_data, y_data)

        # call `next` to get one sample from each tensor:
        xs = next(iter(ds))
        print(str(xs)) # markdown-exec: hide
        ```
        """
        self.tensors = tensors

    def __iter__(self) -> Iterator:
        if len(set([t.size(0) for t in self.tensors])) != 1:
            raise ValueError("Size of first dimension must be the same for all tensors.")

        for idx in cycle(range(self.tensors[0].size(0))):
            yield tuple(t[idx] for t in self.tensors)


def to_dataloader(*tensors: Tensor, batch_size: int = 1, infinite: bool = False) -> DataLoader:
    """Convert torch tensors an (infinite) Dataloader.

    Arguments:
        *tensors: Torch tensors to use in the dataloader.
        batch_size: batch size of sampled tensors
        infinite: if `True`, the dataloader will keep sampling indefinitely even after the whole
            dataset was sampled once

    Examples:

    ```python exec="on" source="above" result="json"
    import torch
    from qadence.ml_tools import to_dataloader

    (x, y, z) = [torch.rand(10) for _ in range(3)]
    loader = iter(to_dataloader(x, y, z, batch_size=5, infinite=True))
    print(next(loader))
    print(next(loader))
    print(next(loader))
    ```
    """
    ds = InfiniteTensorDataset(*tensors) if infinite else TensorDataset(*tensors)
    return DataLoader(ds, batch_size=batch_size)


@singledispatch
def data_to_device(xs: Any, *args: Any, **kwargs: Any) -> Any:
    """Utility method to move arbitrary data to 'device'."""
    raise ValueError(f"Unable to move {type(xs)} with input args: {args} and kwargs: {kwargs}.")


@data_to_device.register
def _(xs: None, *args: Any, **kwargs: Any) -> None:
    return xs


@data_to_device.register(Tensor)
def _(xs: Tensor, *args: Any, **kwargs: Any) -> Tensor:
    return xs.to(*args, **kwargs)


@data_to_device.register(list)
def _(xs: list, *args: Any, **kwargs: Any) -> list:
    return [data_to_device(x, *args, **kwargs) for x in xs]


@data_to_device.register(dict)
def _(xs: dict, *args: Any, **kwargs: Any) -> dict:
    return {key: data_to_device(val, *args, **kwargs) for key, val in xs.items()}


@data_to_device.register(DataLoader)
def _(xs: DataLoader, *args: Any, **kwargs: Any) -> DataLoader:
    return DataLoader(data_to_device(xs.dataset, *args, **kwargs))


@data_to_device.register(DictDataLoader)
def _(xs: DictDataLoader, device: torch_device) -> DictDataLoader:
    return DictDataLoader({key: data_to_device(val, device) for key, val in xs.dataloaders.items()})


@data_to_device.register(InfiniteTensorDataset)
def _(xs: InfiniteTensorDataset, device: torch_device) -> InfiniteTensorDataset:
    return InfiniteTensorDataset(*[data_to_device(val, device) for val in xs.tensors])


@data_to_device.register(TensorDataset)
def _(xs: TensorDataset, device: torch_device) -> TensorDataset:
    return TensorDataset(*[data_to_device(val, device) for val in xs.tensors])
